#This source file belongs to e:doc
#Copyright (C) 1999  Thomas M. Schmickl
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License.
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#See the GNU General Public License for more details.
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package BaseForm;

use strict;

=head1 NAME:

Object: BaseForm

=head1 SYNOPSIS:

use class::BaseForm;

=head1 DIRECTORY/FILE:

The file BaseForm.pm is located in the subdirectory './class'

=head1 DESCRIPTION:

The object BaseForm is the base class for all major forms in e:doc. These are
the DocForm and the PrjForm classes until now.

Al methods that have to be available in all major forms like building up a speedbars and menus,
displaying several kinds of dialogs or message boxes or executing macros are done here.

=head1 STATUS:

Coding: 99% done

Documentation: up-do-date

=head1 ATTRIBUTES:

Do not set/read them directly, use the Get/Set methods instead.

$self->{'ui'} #The barnch that hold ALL GUI elements

$self->{'ui'}->{'top'} #Reference to the Toplevel or Mainwindow that is displayed

$self->{'dialogs'};  #A reference to an array of references to dialog windows that have been cretaed by the Docform object (TagForm, TblForm, etc.)

$self->{'content_changed'} #Flag if the content is changed "*" means changed " " means unchanged

$self->{'filename'} #Current filename

$self->{'filename_short'} #Current filename without path

$self->{'path'} #Current path (without filename)

=head1 METHODS:

=head2 Class Methods:

=head2 Constructors:

=head3 new

$object = BaseForm->new($config_name); 

This constructor creates a new object.

This method is never called directly (BaseForm is a base class) , it is only
called from within the constructors of the derived classes.

Parameters:

(+) $config_name: The name of the Section in edoc.cfg that describes the GUI of the derived from

=cut

###############
# Konstructor #
###############
sub new
{
my $class = shift;
my $config_name = shift;

my $self = {};
bless ($self,$class);

$self->{'ui'}={}; #this branch contains all the widgets that build up the GUI of the derived Form-object
$self->{'dialogs'} = [];
$self->{'content_changed'} = ''; #Unchanged
$self->{'filename'} = '* scratch *'; #No file loaded yet

uiBaseForm::build_ui($self, $config_name);

#Store the object in the global @main::forms array
push (@main::forms, $self);

#The Cleanup-Function &do_destroy  is needed for:
#1.) Asking the user to save unsaved changes in the document
#2.) To remove not only the Toplevel but also the perl object itself
$self->top()->protocol('WM_DELETE_WINDOW',sub{$self->do_destroy()}); #Binding the cleanupfunction to the closing of the window BEFORE the Subwidgets are destroyed

return $self;
}

=head2 Destructors:

=head3 DESTROY; 

The standard automatically called PERL destructor for objects.

=cut

#############################
# Automatic Perl destructor #
#############################
sub DESTROY
{
my $self = shift;

}

=head3 do_destroy

$object ->do_destroy(); 

This destructor is binded to the protocol-message WM_DELETE_WINDOW

=cut

##############
# Destruktor #
##############
sub do_destroy
{
my $self = shift;

#Destroy all still-existing dialog-objects manually
if (defined $self->dialogs())
	{
	foreach my $dlg (@{$self->dialogs()})
		{
		$dlg->do_destroy();
		}
	$self->dialogs( undef );
	}

#Remove this BaseForm object from the global @main::forms array
for (my $i = 0; $i <= $#main::forms; $i++)
	{
	#Get the stored Form-reference there
	my $dlg = $main::forms[$i];
	#If its this DocForm
	if ($dlg == $self)
		{
		#remove it from the array
		splice (@main::forms,$i,1);
		last;
		}
	}

#The default callback for WM_DELETE_WINDOW is not used, because we replaced it with this on_destroy-Function, so we have to destroy the Toplevel manually
$self->top()->destroy();

foreach my $key (keys %{$self})
	{
	delete $self->{$key};
	}

#remove the DocForm-Object itself, after the user closed the window
$self = undef;
}

=head2 Public data methods (=Get/Set methods):

=cut

##############################
### PUBLIC GET/SET METHODS ###
##############################

=head3 top

$ref_toplevel = $object->top(); #gets a reference to the toplevel

$object ->top($ref_toplevel); #sets a reference to the toplevel

=cut

### top ###
sub top
{
my $self=shift;
@_ ? $self->{'ui'}->{'top'} = shift : return $self->{'ui'}->{'top'};

}

=head3 filename

$filename = $object->filename(); #gets the current filename

$object->filename($filename); #sets the current filename

=cut

### filename ###
sub filename
{
my $self=shift;
my $filename = shift;
if (defined $filename)
	{
	$self->{'filename'} = $filename;
	my $filename_short;
	my $path;
	if (not ($filename eq '* scratch *') )
		{
		$filename_short = Utils::extract_filename($filename);
		$path = Utils::extract_path($filename);
		}
	if (defined $filename_short)
		{
		$self->filename_short($filename_short);
		}
	else
		{
		$self->filename_short('');
		}
	if (defined $path)
		{
		$self->path($path);
		}
	else
		{
		$self->path('');
		}
	}
else
	{
	return $self->{'filename'};
	}
}

=head3 filename_short

$filename_short = $object->filename_short(); #gets the current filename without path

$object->filename_short($filename_short); #sets the current filename  without path

=cut

### filename_short ###
sub filename_short
{
my $self=shift;
@_ ? $self->{'filename_short'} = shift : return $self->{'filename_short'};
}

=head3 path

$path = $object->path(); #gets the current path

$object->path($path); #sets the current path

=cut

### path ###
sub path
{
my $self=shift;
@_ ? $self->{'path'} = shift : return $self->{'path'};
}

=head3 ref_filename

$ref_filenamename = $object->ref_filename(); #returns a reference to the current filename (for use with Tk widgets that need a reference, READONLY)


=cut

### ref_filename ###
sub ref_filename
{
my $self=shift;
return \$self->{'filename'};
}


=head3 content_changed

$status = $object->contentt_changed(); #gets the current changing status

$object->content_changed($status); #sets the current changing status

($status is an empty string (='') if the content is unchanged and '*' if the content is changed.)

=cut

### content_changed ###
sub content_changed
{
my $self=shift;
@_ ? $self->{'content_changed'} = shift : return $self->{'content_changed'};
}

=head3 ref_content_changed

$ref_status = $object->ref_content_changed(); #gets a reference to the current changing status

($status is an empty string (='') if the content is unchanged and '*' if the content is changed.)

=cut

### ref_content_changed ###
sub ref_content_changed
{
my $self=shift;
return \$self->{'content_changed'};
}

=head3 status_bar

$status_bar = $object->status_bar(); #gets a reference to the status_bar

$object->status_bar($status_bar); #sets a reference to the status_bar

=cut

### status_bar ###
sub status_bar
{
my $self=shift;
@_ ? $self->{'ui'}->{'bottomframe'} = shift : return $self->{'ui'}->{'bottomframe'};
}


=head3 file_label

$file_label = $object->file_label(); #gets a reference to the file_label in the statusbar

$object->file_label($file_label); #sets a reference to the file_label in the statusbar

=cut

### file_label ###
sub file_label
{
my $self=shift;
@_ ? $self->{'ui'}->{'file_label'} = shift : return $self->{'ui'}->{'file_label'};
}


=head3 progress_bar

$progress_bar = $object->progress_bar(); #gets a reference to the progress_bar

$object->progress_bar($progress_bar); #sets a reference to the progress_bar

=cut

### progress_bar ###
sub progress_bar
{
my $self=shift;
@_ ? $self->{'ui'}->{'progress_bar'} = shift : return $self->{'ui'}->{'progress_bar'};
}

=head3 menu

$menu = $object->menu(); #gets a reference to the menu

$object->menu($menu); #sets a reference to the menu

=cut

### menu ###
sub menu
{
my $self=shift;
@_ ? $self->{'ui'}->{'menu'} = shift : return $self->{'ui'}->{'menu'};
}

=head3 speedbar

$speedbar = $object->speedbar(); #gets a reference to the speedbar

$object->speedbar($speedbar); #sets a reference to the speedbar

=cut

### speedbar ###
sub speedbar
{
my $self=shift;
@_ ? $self->{'ui'}->{'speedbar'} = shift : return $self->{'ui'}->{'speedbar'};
}


=head3 dialogs

$ref_dialogs = $object->dialogs(); #gets an reference to the list (array) of all opened dialog forms that must be destroyed manually

$object->dialogs($ref_dialogs); #sets an reference to the list (array) of all opened dialog forms that must be destroyed manually

=cut

### dialogs ###
sub dialogs
{
my $self=shift;
@_ ? $self->{'dialogs'} = shift : return $self->{'dialogs'};
}

=head2 Other public methods:

=head3 is_content_changed

$flag = $object->is_content_changed(); 

returns 0 if the content is unchanged else returns 1

=cut

###########################
### is_content_changed ###
###########################
sub is_content_changed
{
my $self=shift;
if ($self->content_changed() eq '') 
	{
	return 0;
	}
else
	{
	return 1;
	}
}

=head3 set_content_changed

$object->set_content_changed(); 

sets the content-changed-flag to 'changed' (='*')

=cut
 
############################
### set_content_changed ###
############################
sub set_content_changed
{
my $self=shift;
$self->content_changed('*');
}

=head3 set_content_unchanged

$object->set_content_unchanged(); 

sets the content-changed-flag to 'unchanged' (='')

=cut

##############################
### set_content_unchanged ###
##############################
sub set_content_unchanged
{
my $self=shift;
$self->content_changed('');
}

=head3 show_aboutbox

$object->show_aboutbox(); 

Shows the aboutbox of e:doc.

=cut

#####################
# Show the aboutbox #
#####################
sub show_aboutbox
{
my $self = shift;

$self->top()->Busy();

if (not defined $main::aboutbox_btn)
	{
	$main::aboutbox_progress->packForget();
	$main::aboutbox_btn = $main::border->Button(-text=>'Close', -command=>sub{$main::splash_screen->withdraw; $self->top()->raise()})->pack(-side=>'bottom', -fill=>'x', -expand=>1); 
	}
else
	{
	$main::aboutbox_btn->configure(-command=>sub{$main::splash_screen->withdraw(); $self->top()->raise()} )
	}
$main::splash_screen->deiconify();
$main::splash_screen->raise();

$self->top()->Unbusy();
}


=head3 execute_macro

$return_value = $object->execute_macro($macro_name [, %args]);

This method executes a macro in the actual Form object and returns the SCALAR return value.

Parameters:

(+) $macro_name: The name of the macro to execute

(+) %args: a hash holding the arguments that are passed to the macro (OPTIONAL)

=cut

#######################
### Execute a macro ###
#######################
sub execute_macro
{
my $self = shift;
my $macro_name = shift;

$self->top()->Busy();

my %args;
while (@_) 
	{
	my $key = shift;
	my $value = shift;
	if (defined $key)
		{
		$args{$key} = $value;
		}
	};

my $ref_args = {};
if (defined %args)
	{
	$ref_args = \%args;
	}

my $oneliner = undef;
if ($macro_name =~ /%ONELINER::(.*)%/)
	{
	$oneliner = $1;
	$macro_name = undef;
	}

my $return_value = Macro::execute($macro_name, $self, $ref_args, $oneliner);

$self->top()->Unbusy();

return $return_value;
}

=head3 message

$object->message($message_text, $message_title);

Shows a dialogbox with a message.

Parameters:

(+) $message_text: The text of the message to show

(+) $message_title: The title of the message to show

=cut

##################################
### Ein Dialogfenster anzeigen ###
##################################
sub message
{
my $self = shift;
$self->top()->Busy();

my $message_text = shift;
my $message_title = shift;
$message_title = 'Message:' if !defined $message_title;
$message_text = 'Some text is missing here to be cluefull, please report to edoc@magnet.at' if !defined $message_text;
$self->top()->Dialog(-title=>$message_title, -text=>$message_text)->Show();

$self->top()->Unbusy();
}

=head3 dialog

my $new_value = $object->dialog($message_text, $message_title, $default_value );

Shows a dialogbox and queries an input from the user.

The function returns the value entered by the user

Parameters:

(+) $message_text: The text of the message to show

(+) $message_title: The title of the message to show

(+) $default_value: The value that should be entered by default

=cut

##################################
### Ein Dialogfenster anzeigen ###
##################################
sub dialog
{
my $self = shift;
my $message_text = shift;
my $message_title = shift;
my $default_value = shift;

$self->top()->Busy();

require Tk::DialogBox;
my $value = $default_value;
$value = '' if not defined $value;
$message_title = 'Input:' if  not defined $message_title;
$message_text = 'Some text is missing here to be cluefull, please report to edoc@magnet.at' if  not defined $message_text;
my $dlg = $self->top()->DialogBox(-title=>$message_title,  -buttons=>['Ok']);
my $top = $dlg->Subwidget('top');
my $label = $top->Label(-text=>$message_text, -anchor=>'center', -justify=>'center')->pack(-fill=>'both', -expand=>1);
my $entry = $top->Entry(-textvariable=>\$value)->pack(-fill=>'x', -expand=>1);
$entry->selectionRange(0,'end');
$entry->focus();
$dlg->Show(-global=>1);

$self->top()->Unbusy();

return $value;
}

=head2 Private methods:

=head3 _progress_init

$object->_progress_init();

Initialises the progress bar in the right corner of the bottom status bar.

=cut

##############################
# Progressbar initialisieren #
##############################
sub _progress_init
{
my $self = shift;
my $msg = shift;
$self->progress_bar()->createRectangle(2, 2, 201, 16, '-outline'=>'black', '-tags'=>'Base');
$self->{'progress_msg'} = $msg;
$self->{'progress_percent'} = 0;
$self->_progress_update($msg,0);
}

=head3 _progress_update

$object->_progress_update($msg, $percent);

Redraws the progress bar with the given message and percent status.

Parameters:

(+) $msg: The message that should be displayed in the progress bar

(+) $percent: How many percent of the full width should be drawn as a colored bar.

=cut

###########################
# Progressbar auffrischen #
###########################
sub _progress_update
{
my $self = shift;
my $msg = shift;
my $percent = shift;
my $config = $main::config;
$self->{'progress_msg'} = $msg;
$self->{'progress_percent'} = $percent;
$self->progress_bar()->delete('Rec');
$self->progress_bar()->createRectangle(2,2,($self->{'progress_percent'}+2)*2,16,-fill=>$config->{'UI'}->{'activebackground'},-tags=>'Rec');
$self->progress_bar()->delete('Msg');
$self->progress_bar()->createText(4,16,-anchor=>'sw',-text=>$self->{'progress_msg'},-tags=>'Msg');
$self->progress_bar()->update;
}

=head3 _progress_finish

$object->_progress_finish();

Terminates the progress bar and remove the whole progress bar.

=cut

##########################
# Progressbar abschalten #
##########################
sub _progress_finish
{
my $self = shift;
$self->progress_bar()->delete('Rec');
$self->progress_bar()->delete('Msg');
$self->progress_bar()->delete('Base');
$self->{'progress_msg'} = '';
$self->{'progress_percent'} = 0;
}


1; #So muss package enden


=head1 KNOWN BUGS:

=head1 STILL MISSING:


=cut
