#This source file belongs to e:doc
#Copyright (C) 1999  Thomas M. Schmickl
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License.
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#See the GNU General Public License for more details.
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package NumList;

use strict;
require class::NumLabel;
require Tk::Photo;
=head1 NAME:

Object: NumList

=head1 SYNOPSIS:

use class::NumList;

=head1 DIRECTORY/FILE:

The file NumList.pm is located in the subdirectory './class'

=head1 DESCRIPTION:

The NumList object is the core element of the numbering engines you can define.

It works together with the NumLabel objects which are kept in a sorted list (sorted by linenumber in the text widget)
in the NumList object. The NumList objects does calculate the right numbering the NumLabels should show.

There are different kinds of NumLists used in e:doc:

(+) The NumList for paragraph numbering. [ $docform()->tagman()->par_num_list() ]

(+) The NumLists for numbering Images and Tables. [ $docform()->tagman()->img _num_list() and $docform()->tagman()->tbl_num_list() ] 

(+) Later there will be multiple numlists (defined in 'edoc.obj') you can associate with embeded objects instead of the img_num_list and tbl_num_list.
These will be kept in $docform()->tagman()->num_lists()

(+) Several NumLists for numbered lists and bullet lists kept in $docform()->tagman()->lists()
=head1 STATUS:

Coding: 90% done

Documentation: up-do-date

=head1 ATTRIBUTES:

Do not set/read them directly, use the Get/Set methods instead.

$self->{'do_number'} #if the associated NumLabels should be numbered

$self->{'max_number_depth'} #The maximum numbering depth of the list (Level 3 = '2.3.4')

$self->{'start_number_level'} #At which level the numbering should start

$self->{'tagman'} #A reference to the TagMan object the NumList belongs to (!!!CIRCULAR REFERENCE!!!)

$self->{'bullet_file'} #The filename of the bullet image file that should be used if the NumList shows bullets

...The properties show_bullet, show_number should move from the NumLabel object to the NumList object, because these attributes are valid
for the whole list.

$self->{'bullet_img'} #A reference to the Photo widget that shows the bullet image (Only set if the list is used for a bullet list)

@{$self->{'list'}} #An reference to the sorted array which holds the NumLabel items that are added to this NumList in a sorted list

=head1 METHODS:

=head2 Class Methods:

=head2 Constructors:

=head3 new

my $object = NumList->new($start_number_level, $do_number, $max_number_depth, $tagman , $bullet_file); 

This creates a new object of the class NumList and returns a reference to it.

FIXME: Here we have to check if the given bullet_file really exists an give an error message if it doesn't

Parameters: 

(+) $start_number_level: At what level the numbering should start

(+) $do_number: If the numbering should be performed (To allow On/Of switching of the engines)

(+) $max_number_depth: How far (to what level) the numbering should work.

(+) $tagman: A reference to the TagMan object the NumList belongs to.

(+) $bullet_file: The filename of the bullets image file

=cut

###############
# Konstructor #
###############
sub new
{
my $class = shift;
my $start_number_level=shift; #Anfangs-Level fuer Nummerierung
my $do_number =shift; #Flag, ob ueberhaupt nummeriert wird
my $max_number_depth=shift; #Wie tief nummeriert wird (bis zu welchem Level)
my $tagman =shift; #Zeiger auf das zugehrige TagMan-Objekt 
my $bullet_file = shift; #Ist nur bei Bullet-Listen in Verwendung, sonst undef
my $self = {};
bless ($self,$class);
@{$self->{'list'}}=(); # Array mit Referenzen auf NumLabel-Elementen als Eigenschaft einrichten
$self->do_number( $do_number ); 
$self->max_number_depth( $max_number_depth ); 
$self->start_number_level(  $start_number_level );
$self->tagman( $tagman );
#FIXME: Here we have to check if the file exists an give an error message if it doesn't
$self->bullet_file( Utils::process_filename($bullet_file ) ) if defined $bullet_file ;
if (defined $self->bullet_file())
	{
	$self->bullet_img( $self->tagman()->docform_wnd()->Photo(-file=>$self->bullet_file()) );
	}
print "Object $self created.\n"; #debug
return $self;
}

=head2 Destructors:

=head3 DESTROY 

The standard automatically called PERL destructor for objects.
Unfortionatly it is called very late (when the whole application terminates.)
This is due to the circular cross references between DocForm, TagMan, NumList etc.
I will have to unlink them manually in the closing method of the document window.

=cut

#############################
# Automatic Perl destructor #
#############################
sub DESTROY
{
my $self = shift;
$self->clear();
print "Object $self destroyed.\n"; #debug
}

=head2 Public data methods (=Get/Set methods):

=cut

##############################
### PUBLIC GET/SET METHODS ###
##############################

=head3 tagman

$ref_tagman = $object->tagman(); #gets a reference to the TagMan object

$object ->tagman($ref_tagman); #sets a reference to the TagMan object

=cut

### tagman ###
sub tagman
{
my $self=shift;
@_ ? $self->{'tagman'} = shift : return $self->{'tagman'};
}

=head3 do_number

$do_number = $object->do_number(); #gets the value of the 'do_number' attribute

$object ->do_number($do_number); #sets the value of the 'do_number' attribute

=cut

### do_number ###
sub do_number
{
my $self=shift;
@_ ? $self->{'do_number'} = shift : return $self->{'do_number'};
}

=head3 max_number_depth

$max_number_depth = $object->max_number_depth(); #gets the value of the 'max_number_depth' attribute

$object ->max_number_depth($max_number_depth); #sets the value of the 'max_number_depth' attribute

=cut

### max_number_depth ###
sub max_number_depth
{
my $self=shift;
@_ ? $self->{'max_number_depth'} = shift : return $self->{'max_number_depth'};
}

=head3 start_number_level

$start_number_level = $object->start_number_level(); #gets the value of the 'start_number_level' attribute

$object ->start_number_level($start_number_level); #sets the value of the 'start_number_level' attribute

=cut

### start_number_level ###
sub start_number_level
{
my $self=shift;
@_ ? $self->{'start_number_level'} = shift : return $self->{'start_number_level'};
}

=head3 bullet_file

$filename = $object->bullet_file(); #gets the value of the 'bullet_file' attribute

$object ->bullet_file($filename); #sets the value of the 'bullet_file' attribute

=cut

### bullet_file ###
sub bullet_file
{
my $self=shift;
@_ ? $self->{'bullet_file'} = shift : return $self->{'bullet_file'};
}

=head3 bullet_img

$ref_photo = $object->bullet_img(); #gets a  reference to the Photo widget that holds the bullet picture

$object ->bullet_img($ref_photo); #sets a  reference to the Photo widget that holds the bullet picture

=cut

### bullet_img ###
sub bullet_img
{
my $self=shift;
@_ ? $self->{'bullet_img'} = shift : return $self->{'bullet_img'};
}

=head3 list

@list = $object->list(); #gets a reference to the sorted array of the NumLabels hold by this NumList (READONLY)

=cut

### list ###
sub list
{
my $self=shift;
return $self->{'list'};
}


=head2 Other public methods:

=head3 clear

$object->clear(); 

Clears the list of NumLabels

=cut

#################
# Loescht Liste #
#################
sub clear
{
my $self=shift;
#Deleting the whole List of Numlabels
my $itemcounter=0; #Elementzaehler
foreach my $item (@{$self->list()})  #Durch alle Elemente in der liste gehen
	{
	splice (@{$self->list()},$itemcounter,1); #Element herausnehmen
	}
$itemcounter++; #Elementzaehler erhoehen

if ((defined $#{$self->list()})&&($#{$self->list()}>=0))
	{
	$self->list( undef )
	}

}

=head3 add_numlabel

$object->add_numlabel($ref_numlabel, $do_not_sort); 

Adds a NumLabel to the NumList.

All NumLabel objects are kept in a sorted list (sorted by linenumber inside of the text widget..sort later by col-index too for the case that some NumLabels have the same line number 
(only possible with Numlabels of EmbObj-derived widgets, not with PAR_STYLES and ENV_STYLES) 
To speed up the sorting, I use the always-divide-in-the-middle method which you can only use if you sort one new element into an already sorted list.
To make things easier the last three iterations are done with a normal bubblesort.

Parameters:

(+) $ref_numlabel: A reference of the NumLabel to add

(+) $do_not_sort: If set to 1, the NumLabel is not sorted into the list, its just appended (For document opening time)
Because of the sorting algorithm you can only use $do_not_sort=1 if you know that the added element belongs to the end
of the list, because the sorting algorithm insists of an already sorted list to add new elements into. You can only know
that you can append a new element at the end when you open a document from disk, because in the saved file the styles are kept
sorted by their appearance in the document. Thats why the InOut::file_open() function is the only place where $do_not_sort is set to 1.


=cut

#########################
# Ein Element einfuegen #
#########################
sub add_numlabel
{
# Alle NumLabel Elemente liegen je nach ihrer Lage im Text (Zeilennummern) sortiert in der Liste.
# Da es um Absatzformate geht, die immer mindestens eine ganze Zeile beanspruchen, kann es keine
# 2 Elemente mit gleicher Zeilennummer geben (uff!).
# Um die Einsortierung zu beschleunigen wird mit der Halbierungsmethode gearbeitet, der Code bleibt aber
# einfacher und kuerzer, wen die letzte Lokalisierung sequentiell erfolgt. 2 kurze zusaetzliche Iterationen machen nix aus.
my $self=shift;
my $numlabel_ref = shift; #Referenz auf einzufuegendes NumLabel
my $do_not_sort = shift; #ist 1 wenn Objekt SICHER als letztes in die Liste kommt (beim Datei-Oeffnen), ansonsten 0
my $line = $numlabel_ref->get_current_line(); #Aktuelle Zeilennummer des Numlabels ermitteln
my $list_max=$#{$self->list()}; #Laenge der Liste
my $range_min=0; #Untere Begrenzung des moeglichen Bereiches wo Element einsortiert gehoert
my $range_max=$list_max; #Obere Begrenzung des moeglichen Bereiches wo Element einsortiert gehoert
my $i=int( ($range_max-$range_min) / 2)+$range_min; #Itemnummer fuer Erste ueberpruefung berechnen

# if Label gets not sorted, list is empty yet or last item is smaller then actual one -> add at the end
if (($do_not_sort == 1) or  (not ($list_max >= 0)) or (@{$self->list()}->[$list_max]->get_current_line()<$line)) 
	{
	#Add after the last item
	push (@{$self->list()},$numlabel_ref);
	goto EXIT; #sorry for goto
	}

#If first item is higher than the actual one
if (@{$self->list()}->[0]->get_current_line()>$line)
	{
	#Add in front of the last item
	unshift (@{$self->list()},$numlabel_ref);
	goto EXIT; #sorry for goto
	}

#solange mit der Halbierungsmethode arbeiten, bis Stelle auf 5 Items eingegrenzt wurde
until (($range_max-$range_min)<=5)
	{
	#Halbierungs-Einschraenkung durchfuehren
	my $itemline=@{$self->list()}->[$i]->get_current_line(); #Zeilennummer des Elementes in Liste ermitteln
	if ($line<$itemline) #Wenn einzubauendes Label davor ist
		{
		$range_max=$i; #Element ist niedriger, also kann Obergrenze herabgesetzt werden
		$i=int(($range_max-$range_min)/2)+$range_min; #Itemnummer berechnen
		}
	if ($line>$itemline) #wenn einzubauendes Label danach ist
		{
		$range_min=$i; #Element ist hoeher, also kann Untergrenze heraufgesetzt werden
		$i=int(($range_max-$range_min)/2)+$range_min; #Itemnummer berechnen
		}
	}

#Sequentiell suchen
my $found = 0;
for(my $x=$range_min;$x<=$range_max;$x++)
	{
	my $itemline=@{$self->list()}->[$x]->get_current_line(); #Zeilennummer des Elements ermitteln
	if ($itemline > $line) #Wenn erster hoeher Zeilennummer gefunden wurde 
		{
		splice( @{$self->list()},$x,0,$numlabel_ref); #uebergebene Referenz davor einfuegen
		$found = 1;
		}
	last if $found == 1;
	}
EXIT:
}

=head3 delete_NumLabel

$object->delete_NumLabel($ref_numlabel); 

Remove the given NumLabel from the NumList.

Parameters:

(+) $ref_numlabel: A reference of the NumLabel to remove

=cut

#########################
# Ein Element entfernen #
#########################
sub delete_NumLabel
{
my $self=shift;
my $numlabel_ref = shift; #Referenz auf zu entfernendes NumLabel-Objekt
my $itemcounter=0; #Elementzaehler
foreach my $item (@{$self->list()})  #Durch alle Elemente in der liste gehen
	{
	if ($item == $numlabel_ref) #Wenn Element gefunden wurde 
		{
		splice (@{$self->list()},$itemcounter,1); #Element herausnehmen
		last;
		}
	$itemcounter++; #Elementzaehler erhoehen
	}
$self->renumber();
}

=head3 renumber

$object->renumber(); 

Cause all NumLabels in the NumList to be renumbered.

=cut

########################
# Neu durchnummerieren #
########################
sub renumber
{
my $self=shift;
if ($self->do_number()==1) #wenn Nummerierung eingeschaltet ist
	{
	my $max_num_depth=$self->max_number_depth();
	my $start_level=$self->start_number_level();
	#Maximale Gliederungsebene suchen
	my $max_used_level=1;
	foreach my $item (@{$self->list()})
		{
		if ((($item->level()>$max_used_level)&&($max_num_depth==-1))||(($item->level()>$max_used_level)&&($item->level()<=$max_num_depth))) 
			{
			$max_used_level=$item->level() 
			}
		};
	#Array fuer Zaehlung aufbauen
	my @counter=();
	for (my $i=$start_level;$i<=$max_used_level;$i++)
		{
		$counter[$i]=0;
		}
	#Zaehlung durchfuehren
	foreach my $item (@{$self->list()}) #Durch alle Numlabelelemente gehen
		{
		my $level=$item->level(); #Level des Nummernelementes auslesen
		my $text=''; #Variable fuer Labeltext 
		$counter[$level]++; #Counter fuer diese Ebene erhoehen
		if ($level<$max_used_level) #Alle darueberliegenden Level (=alle niedrigeren Elemente) auf 0 setzen
			{
			for (my $i=$level+1;$i<=$max_used_level;$i++)
				{
				$counter[$i]=0;
				}
			}
		for (my $i=$start_level;$i<=$level;$i++)
			{
			$text=$text.$counter[$i].'.';
			}
		chop($text);#Letzter Punkt weg;
		if ($level>$max_num_depth)
			{
			$item->text('');
			}
		else
			{
			$text=$item->pre_number_text().$text.$item->post_number_text();
			$item->text($text);
			}
		}
	}
}

=head2 Private methods:

nothing here yet.

=cut

1; #So muss ein Package enden

=head1 KNOWN BUGS:

=head1 STILL MISSING:

=cut
