#This source file belongs to e:doc
#Copyright (C) 1999  Thomas M. Schmickl
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License.
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#See the GNU General Public License for more details.
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package TagMan;

use strict;

=head1 NAME:

Object: TagMan

=head1 SYNOPSIS:

use class::TagMan;

=head1 DIRECTORY/FILE:

The file TagMan.pm is located in the subdirectory './class'

=head1 DESCRIPTION:

The TagMan object encapsulates all the functionality that deals with the
different text tags in your document. Each DocForm object (DocForm.pm)
has one associated TagMan object with it. The DocForm object contains the text
of your document (inside of the Text widget) and the TagMan object contains
all of the formatting information. 

First I planned to create a new Text widget which inherits from the original Text widget 
and adds the functionality of the TagMan object, but there are several reasons to 
keep the stuff in 2 different objects.The major reason is that this design allows 
to add something like a document/view concept later. 

The TagMan object with all its asociated objects could be the document (only the pure text info is missing 
there) and the DocForm object could act as a view.But this is a different story...

There are also other object associated with the TagMan object:

(+)Several NumList objects (NumList.pm), which allow the numbering
of paragraphs, images and tables

(+)Always 1 NumList object per 'Numbered list'-environments  that are 
placed in the document.

(+) Always 1 Object of a class derived from EmbObj (EmbObj.pm) per embeded
object in your document.

All tags are created in the private method called '_before_add_tag(...)' and completed in the method '_after_add_tag(...)' which ensures that
the naming of the tags follows a concrete scheme. Tags are named as followed:

'<STYLEGROUP><STYLENAME><DATA>:ID' 

... what means that if you have an environment style (<END_STYLES>)
with the name Itemize (<Itemize>) defined in the file <CODE>edoc.int</CODE>, one tag of this
style could have the name '<ENV_STYLES><Itemize><some info>:12', another one could be named '<ENV_STYLES><Itemize>:54'.
The paragraph style 'Chapter' could be '<PAR_STYLES><Chapter><>:2' and a table object (which belongs
to the style group <EMBOBJ>) could be named <EMBOBJ><TBLOBJ><2||aaaa||bbbb>:34.
All this stuff is very important because the style group and the style could be easily
extracted from the name and the id ensures that every tag has a unique name. This is necessary
because each tag can have multiple arguments and argument values with it.

=head1 STATUS:

Coding: 80% done

Documentation: up-do-date

=head1 ATTRIBUTES:

Do not set/read them directly, use the Get/Set methods instead.

$self->{'tag_counter'} = $start_value; #An internal counter for tagnumbering

$self->{'docform'}  = $ref_docform; #Reference to the corresponding DocForm object

$self->{'tags'}->{'names'} = $ref_namelist; #Reference to an Array of all stored tagnames

$self->{'tags'}->{$tagname}->{'positions'} = $number_of_positions; #How many seperate positions of a tag are stored (

$self->{'tags'}->{$tagname}->{$posnumber}->{'start_index'} = $index1; #Start index of a Position (The first position has number 0)

$self->{'tags'}->{$tagname}->{$posnumber}->{'end_index'} = $index2; #End index of a Position (The first position has number 0)

$self->{'args'}->{$tagname}->{'arg_number'} = $number_of_arguments; #How many arguments are defined to this tags style 

$self->{'args'}->{$tagname}->{$argnumber}->{'name'} = $argumet_name; #Name of the argument (The first argument has number 0)

$self->{'args'}->{$tagname}->{$argnumber}->{'type'} = $argumet_type; #Type of the argument. Possible values: TEXT, NUMBER, MULTILINE, OPTION([item1, item2, ...]), REFERENCE([$stylegroup1.$stlye1.tag_name, $stylegroup1.$stlye1.property_value(property_name)], ...), BOOL

$self->{'args'}->{$tagname}->{$argnumber}->{'value'} = $argumet_value; #The current value of the argument, if the arguments type is 'MULTILINE' then this
value is a reference to an anonymous array that stores the lines

$self->{'objects'}->{$tagname} = ref_object; #Reference to an embeded object that is derived from the class 'EmbObj.pm'

$self->{'env_numlists'}->{$tagname} = ref_numlist; #Reference to an NumList object that manages the NumLabels of an environment style that
is a numbered list or a bullet list.

$self->{'doc_numlists'}->{$name} = ref_numlist; #Reference to an NumList objects that are read from the file 'edoc.int' and used for paragraph and embed-object numbering

=head1 METHODS:

=head2 Class Methods:

=head2 Constructors:

=head3 new

my $object = TagMan->new($ref_docform); 

This creates a new object of the class TagMan and returns a reference to it.

Parameters: 

(+) $ref_docform: A reference to the DocForm object this new TagMan object belongs to. 
    It initialises the internal attributes and creates the NumList object that 
    belong to the new TagMan object.

=cut

###############
# Constructor #
###############
sub new
{
my $class = shift;
my $ref_docform = shift; #Zeiger auf Dokumentfenster
my $self = {};
bless ($self,$class);
### hier werden Eigenschaften initialisiert und eventuel Ereignisse gebunden ###
$self->docform($ref_docform); #Zeiger auf jenes DocForm-Objekt zu dem TagMan-Objekt gehoert
$self->_tag_counter(1); # Eindeutige ID-Nummern fuer Tags
$self->{'tags'}={}; # Hash zum Speichern der Taginformationen
$self->{'doc_args'}={}; # Hash zum Speichern der Taginformationen
$self->{'args'}={}; # Hash zum Speichern der zu den Tags gespeicherten Argumente
$self->{'objects'}={}; # Hash zum Speichern der eingebetten Objekte

$self->create_num_lists();

# Liste mit jenen Format-Properties, die zur Darstellungsaenderung der Tags verwendet werden koennen
$self->{'display_options'}=['font','background','borderwidth','foreground','justify','relief','spacing1','spacing2','spacing3','wrap','offset','lmargin1','lmargin2','overstrike','underline','bgstipple','fgstipple','rmargin','tabs','padx','pady','align','stretch'];


print "Object $self created.\n"; #debug
return $self;
}

=head2 Destructors:

=head3 DESTROY 

The standard automatically called PERL destructor for objects.
Unfortionatly it is called very late (when the whole application terminates.)
This is due to the circular cross references between DocForm, TagMan, NumList etc.
I will have to unlink them manually in the closing method of the document window.

=cut

#############################
# Automatic Perl destructor #
#############################
sub DESTROY
{
my $self = shift;
print "Object $self destroyed.\n"; #debug
}

=head2 Public data methods (=Get/Set methods):

=cut

##############################
### PUBLIC GET/SET METHODS ###
##############################

=head3 docform

$ref_docform = $object->docform(); #gets a reference to the corresponding DocForm object

$object ->docform($ref_docform); #sets a reference to the corresponding DocForm object

=cut

### docform ###
sub docform
{
my $self=shift;
@_ ? $self->{'docform'} = shift : return $self->{'docform'};
}

=head3 docform_wnd

$ref_docform_wnd = $object->docform_wnd(); 

Gets a reference to the Toplevel of the corresponding DocForm object

=cut

###############
# docform_wnd #
###############
sub docform_wnd
{
my $self=shift;
#READONLY ATTRIBUTE
return $self->docform->top();
}

=head3 docform_ui

$ref_docform_ui = $object->docform_ui(); 

Gets a reference to the {'ui'}-branch of the corresponding DocForm object.

=cut

##############
# docform_ui #
##############
sub docform_ui
{
my $self=shift;
#READONLY ATTRIBUTE
return $self->docform()->{'ui'};
}

=head3 docform_doctext

$ref_docform_doctext = $object->docform_doctext(); 

Gets a reference to the text widget inside of the corresponding DocForm object.

=cut

###################
# docform_doctext #
###################
sub docform_doctext
{
my $self=shift;
#READONLY ATTRIBUTE
return $self->docform()->doctext();
}

=head3 doc_args_number

$doc_args_number = $object->doc_args_number(); #gets the number of document arguments

$object->doc_args_number($doc_args_number); #sets the number of document arguments

=cut

### doc_args_number ###
sub doc_args_number
{
my $self=shift;
@_ ? $self->{'doc_args'}->{'arg_number'} = shift : return $self->{'doc_args'}->{'arg_number'};
}

########################
# Other public methods #
########################

=head2 Other public methods:

=head3 set_doc_arg_value

$object->set_doc_arg_value($index, $value); 

Parameter:

(+) $index : The index of the argument

(+) $value : The value to be set

=cut

#########################
### set_doc_arg_value ###
#########################
sub set_doc_arg_value
{
my $self=shift;
my $index = shift;
my $value = shift;
if ($main::internal->{'DOC_STYLES'}->{$self->docform()->current_doc_style()}->{'arg_number'} >= $index )
	{
	$self->{'doc_args'}->{$index}->{'value'} = $value;
	}
else
	{
	print "ERROR: Tried to set value of document-argument number $index, \nbut the current document has not so many arguments defined.\n";
	}

}


=head3 refresh_doc_args

$object->refresh_doc_args();

Refreshes the Arguments that belong to that document_style

=cut

#########################
### Refresh doc_args  ###
#########################
sub refresh_doc_args
{
my $self = shift;
$self->{'doc_args'} = {};
my $doc_style = $self->docform()->current_doc_style();
print "Refreshing arguments for DOC_STYLES::$doc_style\n"; #debug1
$self->_add_arguments('', 'DOC_STYLES', $doc_style);
}

=head3 init_num_lists

$object->init_num_lists();

$object -> init_num_lists(); 

Passes the numbering properties of the current document
into the right properties of the NumList objects and perform a renumbering of all
already existing NumLabels if necessary

=cut

########################################
### NumList-Elemente initialisieren  ###
########################################
sub init_num_lists
{
my $self = shift;

#The right point in the $main::internal hash
my $internal_branch = $main::internal->{'DOC_STYLES'}->{$self->docform()->current_doc_style()};

#SETTING THE DEFAULT VALUES FOR THE NUMLISTS
foreach my $list_name (keys %{$main::internal->{'DOC_NUMLISTS'}})
	{
	#Get the start level for numbering
	my $start_number_level = $main::internal->{'DOC_NUMLISTS'}->{$list_name}->{'start_number_level'};
	#Get the on/off state 
	my $do_number = $main::internal->{'DOC_NUMLISTS'}->{$list_name}->{'do_number'};
	#Get the maximum number depth
	my $max_number_depth = $main::internal->{'DOC_NUMLISTS'}->{$list_name}->{'max_number_depth'};
	#Setting the default properties of the NumList
	$self->{'doc_numlists'}->{$list_name}->start_number_level($start_number_level);
	$self->{'doc_numlists'}->{$list_name}->max_number_depth($max_number_depth);
	$self->{'doc_numlists'}->{$list_name}->do_number($do_number);
	#Do not renumber here->time is money
	}

#NOW OVERRIDING WITH THE VALUES SET IN THE DOC_STYLES-DEFINITIONS
#Get the start level for numbering
my $start_number_level = $internal_branch->{'start_number_level'};
#Split up the '||'-separated list
my @list = split (/\|\|/, $start_number_level);
#Iterate overvthe items
foreach my $term (@list)
	{
	#Extract the names of the NumLists and the values EXAMPLE: 'PAR(1)'
	if ($term =~ /(.+?)\((\d+?)\)/o)
		{
		my $list_name = $1;
		my $level = $2;
		#If the NumList exists
		if (exists $self->{'doc_numlists'}->{$list_name} and defined $self->{'doc_numlists'}->{$list_name})
			{
			if ((defined $level) and (not($level eq '')))
				{
				$self->{'doc_numlists'}->{$list_name}->start_number_level($level);
				}
			}
		}
	}

#Get the maximum level for numbering
my $max_number_depth = $internal_branch->{'max_number_depth'};
#Split up the '||'-separated list
@list = split (/\|\|/, $max_number_depth);
#Iterate overvthe items
foreach my $term (@list)
	{
	#Extract the names of the NumLists and the values EXAMPLE: 'PAR(4)'
	if ($term =~ /(.+?)\((\d+?)\)/o)
		{
		my $list_name = $1;
		my $depth = $2;
		#If the NumList exists
		if (exists $self->{'doc_numlists'}->{$list_name} and defined $self->{'doc_numlists'}->{$list_name})
			{
			if ((defined $depth) and (not($depth eq '')))
				{
				$self->{'doc_numlists'}->{$list_name}->max_number_depth($depth);
				}
			}
		}
	}

#Get the on/off state of the NumList
my $do_number = $internal_branch->{'do_number'};
#Split up the '||'-separated list
@list = split (/\|\|/, $do_number);
#Iterate overvthe items
foreach my $term (@list)
	{
	#Extract the names of the NumLists and the values EXAMPLE: 'PAR(4)'
	if ($term =~ /(.+?)\((\d+?)\)/o)
		{
		my $list_name = $1;
		my $do_it = $2;  #1 means yes, 0 means no
		#If the NumList exists
		if (exists $self->{'doc_numlists'}->{$list_name} and defined $self->{'doc_numlists'}->{$list_name})
			{
			if ((defined $do_it) and (not($do_it eq '')))
				{
				$self->{'doc_numlists'}->{$list_name}->do_number($do_it);
				}
			}
		}
	}
#Renumber the lists if neccessary
foreach my $list_name (keys %{$self->{'doc_numlists'}})
	{
	if (exists $self->{'doc_numlists'}->{$list_name} and defined exists $self->{'doc_numlists'}->{$list_name})
		{
		if ($self->{'doc_numlists'}->{$list_name}->do_number() == 1) 
			{
			$self->{'doc_numlists'}->{$list_name}->renumber();
			}
		}
	}

}

=head3 create_num_lists

$object->create_num_lists();

$object -> Create_num_lists(); 

Creates all NumLists that are defined in 'edoc.int' with the default parameters

=cut

############################
### Create all NumLists  ###
############################
sub create_num_lists
{
my $self = shift;
foreach my $name (keys %{$main::internal->{'DOC_NUMLISTS'}})
	{
	#Get the start level for numbering
	my $start_number_level = $main::internal->{'DOC_NUMLISTS'}->{$name}->{'start_number_level'};
	#Get the on/off state 
	my $do_number = $main::internal->{'DOC_NUMLISTS'}->{$name}->{'do_number'};
	#Get the maximum number depth
	my $max_number_depth = $main::internal->{'DOC_NUMLISTS'}->{$name}->{'max_number_depth'};
	#Create the numlist
	$self->{'doc_numlists'}->{$name}=NumList->new($start_number_level, $do_number, $max_number_depth, $self, undef);
   $self->{'doc_numlists'}->{$name}->renumber();
	}
}


=head3 read_from_widget

$object->read_from_widget($source_text_widget);

Read all tags from the source text widget into the $self->{'tags'} branch.

=cut

#######################################
# Tags aus Textwidget heraus auslesen #
#######################################
sub read_from_widget
{
my $self = shift; #Zeiger auf Objekt
my $text_widget = shift; #Quell-Widget
$self->_clear_tags(); #Tagliste loeschen
my @names = $text_widget->tagNames(); #Tagnamen auslesen
$self->{'tags'}->{'names'}=\@names; #Tagnamen  in Liste speichern
foreach my $name (@names)
	{
	if ($name =~ /^sel$/o) {next}; #Diese Tags ignorieren
	my @ind = $text_widget->tagRanges($name); #Tagindizes auslesen
	my $pos_counter=0; #Positionszaehler (Startet mit 0 !!!!)
		for (my $i = 0; $i <= $#ind; $i=$i+2) #In 2er-Schritten hindurch 	
		{
		$self->{'tags'}->{$name}->{'positions'}=$pos_counter; #Jedesmal Anzahl der Positionen erhoehen
		my $start_index = $ind[$i]; #Anfangsindex der aktuellen Position
		my $end_index = $ind[$i+1];  #Endindex der aktuellen Position
		$self->{'tags'}->{$name}->{$pos_counter}->{'start_index'}=$start_index; #In die Hash damit
		$self->{'tags'}->{$name}->{$pos_counter}->{'end_index'}=$end_index; #In die Hash damit
		$pos_counter++; #Positionszaehler erhoehen
		}
	}
}

=head3 copy_text_widget

$object->copy_text_widget($source_text_widget, $target_source_widget, $index1, $index2);

Copies all text info and tag info from one text widget into another one.

Parameters:

(+) $source_widget: A Reference to the text widget to copy

(+) $target_widget: A Reference to the text widget to copy into

(+) $index1, $index2: The start and the end index of the range to copy (Both should be '-1' if the whole widget should be copied)



=cut

##############################################
# Tags von einem Widget zum anderen kopieren #
##############################################
sub copy_text_widget
{
my $self = shift; #Zeiger auf Objekt
my $source_widget = shift; #Zeiger auf Quell-Textwidget 
my $target_widget = shift; #Zeiger auf Ziel-Textwidget 
my $index1 = shift; #is -1 id the whole text widget should be copied
my $index2 = shift; #is -1 id the whole text widget should be copied

$self->read_from_widget($source_widget); #Aus Quell-Textwidget auslesen
$target_widget->delete('1.0','end');
my @old_tags = $target_widget->tagNames();
foreach my $old_tag (@old_tags)
	{
	$target_widget->tagDelete($old_tag);
	}

my $startindex;
my $endindex;

my $startline;
my $startcol;
my $endline;
my $endcol;

if ((not defined $index1 and not defined $index2) or ($index1 == -1 and $index2 == -1))
	{
	$startline = 1;
	$startcol = 0;
	$endindex = $source_widget->index('end');
	$startindex = '1.0';
	if ($endindex =~ /(\d+)\.(\d+)/o)
		{
		$endline = $1 - 1;
		$endindex = $source_widget->index("$endline.end");
		if ($endindex =~ /(\d+)\.(\d+)/o)
			{
			$endcol = $2;
			}
		}
	}
else
	{
	$startindex = $source_widget->index($index1);
	$endindex = $source_widget->index($index2);
	if ($startindex =~ /(\d+)\.(\d+)/o)
		{
		$startline = $1;
		$startcol = $2;
		}
	if ($endindex =~ /(\d+)\.(\d+)/o)
		{
		$endline = $1;
		$endcol = $2;
		}
	}

my @text; #An array for holding the text to copy
my $targetline = 0;
for (my $line = $startline; $line<=$endline;$line++)
	{
	$targetline++;
	my $line_text ='';
	my @tag_names;
	my @tag_index;
	my $end_of_this_line;
	my $start_of_this_line;
	if ($line == $endline){$end_of_this_line=$endcol}
	else
		{
		$end_of_this_line = $source_widget->index("$line.end");
		$end_of_this_line =~ s/^(\d+)\.//o ;
		}
	if ($line == $startline){$start_of_this_line=$startcol} else {$start_of_this_line = 0};
	my $targetcol=-1;
	for (my $col=$start_of_this_line; $col<=$end_of_this_line; $col++)
		{
		$targetcol++;
		my $char=$source_widget->get("$line.$col");
		$char = ' ' if (not defined $char or $char eq '');
		$line_text .= $char;
		}
	push (@text, $line_text);
	}


for (my $i=0 ; $i<=$#text; $i++)
	{
	$target_widget->insert('end',$text[$i]);
	}

my @tagnames = $source_widget->tagNames();
foreach my $tag (@tagnames)
	{
	my @range = $source_widget->tagRanges($tag);
	while ($#range>=0)
		{
		my $i_start = shift(@range);
		my $i_end = shift(@range);
		#Switch indexes if they are upside-down
		if ($source_widget->compare($i_start,'>',$i_end))
			{
			my $mem = $i_start;
			$i_start = $i_end;
			$i_end = $i_start;
			}
		#clip them if neccessary
		next if ($source_widget->compare($i_end,'<',$startindex) or $source_widget->compare($i_start,'>',$endindex));
		$i_start = $startindex if ($source_widget->compare($i_start,'<',$startindex)); 
		$i_end = $endindex if ($source_widget->compare($i_end,'>',$endindex)); 
		#calculate them to the target_widget
		my $i_start_col;
		my $i_start_line;
		my $i_end_col;
		my $i_end_line;
		if ($i_start =~ /^(\d+)\.(\d+)/o)
			{
			$i_start_line = $1;
			$i_start_col = $2;
			}
		if ($i_end =~ /^(\d+)\.(\d+)/o)
			{
			$i_end_line = $1;
			$i_end_col = $2;
			}
		my $t_start = ($i_start_line - $startline+1).".".($i_start_col-$startcol);
		my $t_end;
		if ($startline==$endline)
			{
			$t_end = ($i_end_line - $startline+1).".".($i_end_col-$startcol);
			}
		else
			{
			$t_end = ($i_end_line - $startline+1).".".$i_end_col;
			}
		$target_widget->tagAdd($tag, $t_start, $t_end);
		}
	}

if ($self->docform()->tag_doctor_on() == 1)
	{
	my @ranges = $target_widget->tagRanges('TAG_DOC');
	while ($#ranges>=0)
		{
		my $start_index= shift(@ranges); 
		my $end_index = shift(@ranges);
		$target_widget->delete($start_index, $end_index);
		@ranges = $target_widget->tagRanges('TAG_DOC');
		}
	}

my @tags = $target_widget->tagNames();
foreach my $tag (@tags)
	{
	if ($tag =~ /^<NUM>/o  or $tag =~ /^<LIST_NUM>/o)
		{
		my @range =$target_widget->tagRanges($tag);
		while ($#range>=0)
			{
			my $start_index = shift(@range);
			my $end_index = shift(@range);
			$target_widget->delete($start_index, $end_index);
			}
		}
	}
}

=head3 sort_tags_by_stylegroup

@taglist = $object->sort_tags_by_stylegroup($ref_text_widget);

Sorts all tags in the following order

CHAR_SYTYLES,COMMANDS, EMBOBJ,ENV_STYLES, PAR_STYLES, CHAR_SYTYLES

Parameters:

(+) $ref_text_widget: A reference to the textwidget that holds the tags to sort



=cut

#########################################################################################################
# Liefert eine sortierte Tagliste in der Reihenfolge ENV_STYLES,PAR_STYLES,CHAR_STYLES,COMMANDS,OBJECTS #
#########################################################################################################
sub sort_tags_by_stylegroup
{
my $self = shift; #Zeiger auf Objekt
my $textwidget = shift;
my @nums;
my @pars;
my @chars;
my @envs;
my @objects; #This means <COMMAND>,<EMBOBJ> and all other Objects that use the space of one character in the textwidget

my @names=$textwidget->tagNames();

foreach my $name (@names)
	{
	if (($name =~ /^<NUM>/o)||($name =~/^<LIST_NUM>/o)||($name =~/^TAG_DOC/o)) {push(@nums, $name)}
	elsif ($name =~ /^<PAR_STYLES>/o) {push(@pars, $name)}
	elsif ($name =~ /^<CHAR_STYLES>/o) {push(@chars, $name)}
	elsif ($name =~ /^<ENV_STYLES>/o) {push(@envs, $name)}
	elsif (($name =~ /^<COMMANDS>/o)||($name =~ /^<EMBOBJ>/o)) {push(@objects, $name)}
	}

@names = ();
push (@names,@nums);
push (@names,@objects);
push (@names,@chars);
push (@names,@envs);
push (@names,@pars);
return @names;
}


=head3 add_par_tag

$tagname = $object->add_par_tag($style, $index1, $index2, $do_not_sort, $allow_add_tagdoc);

Adds a <PAR_STYLES> tag to the textwidget and the $self->{'tags'} branch. All defined
arguments are also initialised in the $self->{'args'} branch. If this paragraph style
is numbered, there will be a NumLabel created automatically and added to the NumList object
this style is associated with. This NumLabel will be automatically sorted and numbered
if $do_not_sort is unequal to 1.

The tags and arguments are created via the private methods '_before_add_tag' and '_after_add_tag'.

The method returns the full tagname of the newly created tag. 

Arguments:

(+) $stlye: The name of the style to add

(+) $index1, $index2: The start index and the end index of the tag range

(+) $do_not_sort: If this flag is set to 1 the possibly automatically created NumLabel should not be sorted and renumbered
to save time on document opening when many tags are created and one renumbering at the end is enough. This is why $self->renumber()
has to be called at the end of those 'file_open' functions.

(+) $allow_add_tagdoc: If set to 0 or undef, do not add TagDoc-objects (disturbing at loading/paste-time)

=cut

#############################
# <PAR_STYLES>-Tag erzeugen #
#############################
sub add_par_tag
{
my $self = shift; #Zeiger auf self-Objekt
my $style = shift; #Absatzformat-Name
my $index1 = shift; #Begin-Index
my $index2 = shift; #End-Index
my $do_not_sort = shift; #Ist 1 wenn Nummernlabel nicht einsortiert werden soll (Datei oeffnen)
my $allow_add_tagdoc = shift;
$allow_add_tagdoc = 0 if not defined $allow_add_tagdoc;

my $text_widget = $self->docform_doctext();
my $internal_branch = $main::internal->{'PAR_STYLES'}->{$style};

#Calculating linenumbers
$index1 = $text_widget->index($index1);
$index2 = $text_widget->index($index2);
(my $startline, my $startcol) = $index1 =~ /^(\d+?)\.(\d+?)$/o;
(my $endline, my $endcol) = $index2 =~ /^(\d+?)\.(\d+?)$/o;

#Tag erzzeugen mit allem was dran haengt
my $tag_name = $self->_before_add_tag('PAR_STYLES',$style,$index1,$index2,$do_not_sort);


my $any_changes = 0; #Flag
if (exists $internal_branch->{'parent_env_styles'} and 
	defined $internal_branch->{'parent_env_styles'} and 
	(not ($internal_branch->{'parent_env_styles'} eq '')))
	{
	my $parent_env_style= $internal_branch->{'parent_env_styles'};
	#Get styles into the @parent_env_styles-array
	my @parent_env_styles = split(/\|\|/, $internal_branch->{'parent_env_styles'});
	
	#this paragraph style belongs to an environment style that has to be there before
	my $found=0; #flag
	for (my $i = 0; $i <= $endcol; $i++) #Iterate over the cols of the line
		{
		my @tagnames = $text_widget->tagNames("$startline.$i");  #Get the existing tagnames of the position
		foreach my $a_tagname (@tagnames) #Iterate over the tagnames
			{
			foreach my $parent_style (@parent_env_styles) #Iterate over all styles in the @parent_env_styles-array
				{
				if ($a_tagname =~ /^<ENV_STYLES><$parent_style>/o ) #Test for the searched ENV_STYLES style
					{
					$found = 1; #Set the flag
					last; #Finnish the loop
					}
				last if $found == 1;
				}
			last if $found == 1;
			}
		last if $found == 1; #Finish this loop also if already found
		}
	if (($found==0) and ($internal_branch->{'only_in_parent'} == 1)) #If the ENV_STYLES-styles wasnt there
		{
		$parent_env_style =~ s/\|\|/, /go ;
		#Show an error-message
		$self->docform()->message("This paragaph style could only be applied into \nexisting tags of these environment-styles: \"$parent_env_style\"", "ERROR: $tag_name");
		$text_widget->tagRemove($tag_name,$index1,$index2); #Remove the already placed tag
		goto EXIT; #Jump out of the whole sub
		}
	}
#Wenn dieses Absatzformat nummeriert wird
if ($internal_branch->{'do_number'} == 1)
	{
	# NUMMERNLABEL ERZEUGEN #
	#Eigenschaften in Variablen schreiben
	my $level = $internal_branch->{'level'}; # Eigenschaft level setzen
	my $num_list;
	if (exists $internal_branch->{'connect_to_doc_numlist'} and
		defined $internal_branch->{'connect_to_doc_numlist'} and
		(not($internal_branch->{'connect_to_doc_numlist'} eq '' )))
		{
		my $num_list_name = $internal_branch->{'connect_to_doc_numlist'};
		$num_list = $self->{'doc_numlists'}->{$num_list_name}; #A reference to the NumList the style is connected to
		
		my $number = 0;
		if ($tag_name =~ /^<(.+?)><(.+?)><(.*?)>:(\d+?)$/ )
			{
			$number = $4;
			}
		my $window_tag = "<NUM><".$style."><>:$number"; # Namen des Markierungstags uebergeben
		my $num_pre_number_text = $internal_branch->{'pre_number_text'}; # Eigenschaft pre_number_text setzen
		my $num_post_number_text = $internal_branch->{'post_number_text'}; # Eigenschaft post_number_text setzen
		my $num_font=$internal_branch->{'num_font'};
		$num_font=$internal_branch->{'font'} if $num_font eq '';
		my $num_borderwidth=$internal_branch->{'num_borderwidth'};
		my $num_background=$internal_branch->{'num_background'};
		my $num_foreground=$internal_branch->{'num_foreground'};
		my $num_relief=$internal_branch->{'num_relief'};
		my $num_show_number = $internal_branch->{'num_show_number'};
		my $num_show_bullet = $internal_branch->{'num_show_bullet'};
		#Nummernlabel erzeugen
		my $par_number_label = $text_widget->NumLabel(-numlist=>$num_list, 
																	-level=>$level, 
																	-pre_number_text=>$num_pre_number_text, 
																	-post_number_text=>$num_post_number_text, 
																	-text_widget=>$text_widget, 
																	-associated_tag=>$tag_name, 
																	-relief=>$num_relief,  
																	-font=>$num_font,
																	-background=>$num_background,
																	-foreground=>$num_foreground,
																	-borderwidth=>$num_borderwidth,
																	-show_number=>$num_show_number, 
																	-show_bullet=>$num_show_bullet);
		#Fenster in Textwidget erzeugen und Nummernlabel dort einbetten
		my $absnummerwindow = $text_widget->windowCreate($index1, -window=>$par_number_label);
		#Tag um Nummernlabel herum erzeugen
		$text_widget->tag("add",$window_tag,$index1,"$index1 + 1 chars"); #Eingebettets Nummernlabel mit einem <Num>-Tag markieren um es von den <PAR_STYLES>-Tags unterscheiden zu koennen (Hat aber die selbe Nummer wie <PAR_STYLES>-Label!!!)
		#Tagname an NumLabel-Objekt uebergeben
		$par_number_label->window_tag( $window_tag );
		#NumLabel-Objekt fertig initialisieren
		$par_number_label->on_create($do_not_sort);
		$any_changes = 1;
		}
	}

if ($any_changes == 1)
	{
	# Re-apply the tag, because there have been things inserted at the begin & the end
	$text_widget->tagRemove($tag_name,"$startline.0","$endline.end");
	$text_widget->tagAdd($tag_name,"$startline.0","$endline.end");
	}

#Do the all the rest with this new tag
$self->_after_add_tag($tag_name, 'PAR_STYLES', $style, $allow_add_tagdoc);

#Tagnamen zurueckliefern
return $tag_name;
}

=head3 add_env_tag

$tagname = $object->add_env_tag($style, $index1, $index2, $allow_add_tagdoc);

Adds a <ENV_STYLES> tag to the textwidget and the $self->{'tags'} branch. All defined
arguments are also initialised in the $self->{'args'} branch. If this environment style
is a sort of list, there will be created a seperate NumList object to handle this list and 
a reference to it will be stored in $self->{'env_numlists'}->{'tagname'}.
There will be a NumLabels created automatically and added to the NumList object
stored in $self->{'env_numlists'}->{'tagname'}.

The tags and arguments are created via the private method '_before_add_tag' and '_after_add_tag'.

The method returns the full tagname of the newly created tag. 

Arguments:

(+) $stlye: The name of the style to add

(+) $index1, $index2: The start index and the end index of the tag range

(+) $allow_add_tagdoc: If set to 0 or undef, do not add TagDoc-objects (disturbing at loading/paste-time)

=cut

#############################
# <ENV_STYLES>-Tag erzeugen #
#############################
sub add_env_tag
{
my $self = shift; #Zeiger auf self-Objekt
my $style = shift; #Absatzformat-Name
my $index1 = shift; #Begin-Index
my $index2 = shift; #End-Index
my $allow_add_tagdoc = shift;
$allow_add_tagdoc = 0 if not defined $allow_add_tagdoc;

my $text_widget = $self->docform_doctext();

my $internal_branch = $main::internal->{'ENV_STYLES'}->{$style};

#Calculating linenumbers
$index1 = $text_widget->index($index1);
$index2 = $text_widget->index($index2);
(my $startline, my $startcol) = $index1 =~ /^(\d+?)\.(\d+?)$/o;
(my $endline, my $endcol) = $index2 =~ /^(\d+?)\.(\d+?)$/o;


#Tag erzzeugen mit allem was dran haengt
my $tag_name = $self->_before_add_tag('ENV_STYLES',$style,$index1,$index2,0);

my $do_bind_enter = 0; #Flag
my $any_changes = 0; #Flag

if (exists $internal_branch->{'do_number_subitems'} and
	defined $internal_branch->{'do_number_subitems'} and
	$internal_branch->{'do_number_subitems'} == 1)
	{
	$any_changes = 1;
	$self->_remove_list_items($tag_name,-1,-1);# Remove old List_Nums if there
	my $bullet_file = Utils::process_filename($internal_branch->{'subitems_bullet_file'});
	$self->{'env_numlists'}->{$tag_name} = NumList->new(1, 1, 1,$self,$bullet_file);
	#Beim anlegen auch gleich alle Items erzeugen ?
	my $auto_add_subitems = $internal_branch->{'auto_add_subitems'};
	if ($auto_add_subitems==1)
		{
		#Alle Items erzeugen
		$self->_build_list_items($tag_name,-1,-1);
		#Enter-Event an Ereignis binden
		$do_bind_enter = 1;
		}
	}

if ($any_changes == 1)
	{
	# Re-apply the tag, because there have been things inserted at the begin & the end
	$text_widget->tagRemove($tag_name,"$startline.0","$endline.end");
	$text_widget->tagAdd($tag_name,"$startline.0","$endline.end");
	}

#If the <KeyRelease-Return> should be bound to a callback
if ($do_bind_enter == 1)
	{
	#Binden
	$text_widget->tagBind($tag_name,'<KeyRelease-Return>',sub{$self->_return_pressed($tag_name)});
	}

#Do the all the rest with this new tag
$self->_after_add_tag($tag_name, 'ENV_STYLES', $style, $allow_add_tagdoc);

#Tagnamen zurueckliefern
return $tag_name;
}

=head3 add_char_tag

$tagname = $object->add_char_tag($style, $index1, $index2, $allow_add_tagdoc);

Adds a <CHAR_STYLES> tag to the textwidget and the $self->{'tags'} branch. All defined
arguments are also initialised in the $self->{'args'} branch.

The tags and arguments are created via the private method '_before_add_tag' and '_after_add_tag'.

The method returns the full tagname of the newly created tag. 

Arguments:

(+) $stlye: The name of the style to add

(+) $index1, $index2: The start index and the end index of the tag range

(+) $allow_add_tagdoc: If set to 0 or undef, do not add TagDoc-objects (disturbing at loading/paste-time)

=cut

##############################
# <CHAR_STYLES>-Tag erzeugen #
##############################
sub add_char_tag
{
my $self = shift; #Zeiger auf self-Objekt
my $style = shift; #Absatzformat-Name
my $index1 = shift; #Begin-Index
my $index2 = shift; #End-Index
my $allow_add_tagdoc = shift;
$allow_add_tagdoc = 0 if not defined $allow_add_tagdoc;

my $text_widget = $self->docform_doctext();
#Create the new char tag
my $tag_name = $self->_before_add_tag('CHAR_STYLES',$style,$index1,$index2,0);

#Some additional code might come here...

#Do the all the rest with this new tag
$self->_after_add_tag($tag_name, 'CHAR_STYLES', $style, $allow_add_tagdoc);

#Tagnamen zurueckliefern
return $tag_name;
}

=head3 add_command_tag

$tagname = $object->add_command_tag($command, $index1);

Adds a <COMMANDS> tag to the textwidget.

The tags and arguments are created in this method (not in '_before_add_tag').

The method returns the full tagname of the newly created tag. 

The 'command' is an object of the class 'ComObj.pm' which is an derived Widget and embeded into the 
text widget. This widget uses the space of 1 character in the text widget and this space is tagged with
the created tag in this method.

As noted in the description of '_before_add_tag', the process is quite different to the process used
for text tags.

Arguments:

(+) $command: The name of the comand style to add

(+) $index1: The start index of the embeded ComObj object

=cut

#####################
# Commando erzeugen #
#####################
sub add_command_tag
{
my $self = shift; #Zeiger auf self-Objekt
my $command = shift; #Kommando-Name
my $index1 = shift; #Einfuege-Index
my $text_widget = $self->docform_doctext();

my $index2 = $text_widget->index("$index1  + 1 chars");

# Darstellungseigenschaften an neu erzeugtes Tag uebergeben
my @comobj_options = ('font','background','borderwidth','foreground','justify','relief','wrap','overstrike','underline');
my %obj_options; #Hash fuer Darstellungsoptionen
my $was_obj_option=0; #Flag

for (my $i=0;$i<= $#comobj_options;$i++) #Durch alle definierten Optionen zur Tagdarstellung gehen
	{
	my $option=$comobj_options[$i]; #Eine Option auslesen
	if (exists($main::internal->{'COMMANDS'}->{$command}->{$option})) #Wenn diese in INTERNAL-Struktur definiert ist
		{
		$was_obj_option=1; #Flag setzen
		$obj_options{"-$option"}=$main::internal->{'COMMANDS'}->{$command}->{$option}; #In Options-Hash eintragen
		}
	}

# Eingebettets Window erzeugen
my $command_obj = $text_widget->ComObj( '-docform'=>$self->docform(), -command=>$command, %obj_options);

#Eingebettetes Window erzeugen
my $command_window = $text_widget->windowCreate($index1, -window=>$command_obj, -stretch=>0, -align=>'center');

#Create the tag around the ComObj
my $tag_name = $self->_before_add_tag('COMMANDS', $command, $index1, $index2);

#Set the -window_tag option of the ComObj
$command_obj->window_tag( $tag_name );

#Do the all the rest with this new tag
$self->_after_add_tag($tag_name, 'COMMANDS', $command, 0);

#Finish the initialisation of the ComObj
$command_obj->on_create();

#Tagnamen zurueckliefern
return $tag_name;
}

=head3 add_embobj_tag

$tagname = $object->add_embobj_tag($classname, $index1, $data, $do_not_sort_numlabel);

Adds a <EMBOBJ> tag to the textwidget.

The tags and arguments are created in this method (not in '_before_add_tag').

The method returns the full tagname of the newly created tag. 

The 'EMBOBJ' is an object derived from the class 'EmbObj.pm' which is an derived Widget and embeded into the 
text widget. This widget uses the space of 1 character in the text widget and this space is tagged with
the created tag in this method.

As noted in the description of '_before_add_tag', the process is quite different to the process used
for text tags.

Arguments:

(+) $classname: The classname of the EmbObj-derived widget to embed. there has to be a file
with the name {classname}.pm in the ./class-directory!

(+) $data : The data that is needed for the EmbObj, is '' if created from scratch

(+) $index1: The start index of the embeded EmbObj object

(+) $do_not_sort_numlabel: Prevent renumbering of the associated numlist at loading time when set to 1

=cut

####################
# Create an EmbObj #
####################
sub add_embobj_tag
{
my $self = shift; #Zeiger auf self-Objekt
my $classname = shift; #Kommando-Name
my $data = shift; #The data that belongs to the EmbObj (from loading/pasting, else '' )
my $index1 = shift; #Einfuege-Index
my $do_not_sort_numlabel = shift; #At loadingtime, there is no need to sort them every time a object is created

$data = '' if not defined $data;

my $text_widget = $self->docform_doctext();
my $top_win = $self->docform()->top();

my $index2 = $text_widget->index("$index1 + 1 chars");

my $internal_branch = $main::internal->{'EMBOBJ'}->{$classname}; #shortcut

# Array with possible display options
my @embobj_options = ('font','background','foreground','relief','pre_number_text','post_number_text','do_number');
my %used_display_options; #Hash for used options
for (my $i=0;$i<= $#embobj_options;$i++) #Iterate over the options
	{
	my $option = $embobj_options[$i]; #Get an option name
	if (exists($internal_branch->{$option})) #If this option is defined in 'edoc.int'
		{
		$used_display_options{"-$option"} = $internal_branch->{$option}; #Store it in the hash
		}
	}

#Getting the numlist
if (exists $internal_branch->{'connect_to_doc_numlist'} and defined $internal_branch->{'connect_to_doc_numlist'})
	{
	my $num_list_name = $internal_branch->{'connect_to_doc_numlist'};
	if (exists $self->{'doc_numlists'}->{$num_list_name} and defined $self->{'doc_numlists'}->{$num_list_name})
		{
		$used_display_options{'-numlist'} = $self->{'doc_numlists'}->{$num_list_name};
		}
	}

#Storing the data there
$used_display_options{'-data'} = $data;

my %obj_config;
foreach my $option (keys %{$internal_branch})
	{
	if (defined $internal_branch->{$option})
		{
		if ($option =~ /^_/ )
			{
			$obj_config{$option} = $internal_branch->{$option};
			}
		}
	}


# Create the widget (HERE COMES THE MAGIC!!!!)
require $classname.".pm";
my $embobj = $text_widget->$classname(
													'-docform'=>$self->docform(), 
													-classname=>$classname, 
													-do_not_sort_numlabel => $do_not_sort_numlabel,
													-obj_config => \%obj_config,
													%used_display_options
												);

#Embed it
my $command_window = $text_widget->windowCreate($index1, -window=>$embobj, -stretch=>0, -align=>'center');

#Create the tag around the EmbObj
my $tag_name = $self->_before_add_tag('EMBOBJ',$classname,$index1,$index2);

#Store a reference to it in the {'objects'} branch of the TagMan
$self->{'objects'}->{$tag_name} = $embobj;

#Set the -window_tag option of the EmbObj
$embobj->window_tag( $tag_name );

#Get the backend definition for this object
my $ref_backend = $self->docform()->backend()->{'EMBED_OBJECTS'}->{$classname};

#Pass the output definition to the object
$embobj->set_backend($ref_backend);

#Do the all the rest with this new tag
$self->_after_add_tag($tag_name, 'EMBOBJ', $classname, 0);

#Finish the initialisation of the EmbObj
$embobj->on_create();

#Tagnamen zurueckliefern
return $tag_name;
}


=head3 delete_par_tags_in_area

$object->delete_par_tags_in_area($index1, $index2);

Removes all <PAR_STYLES> tag from the textwidget and the $self->{'tags'} branch. All defined
arguments are also deleted in the $self->{'args'} branch.

The tags and arguments are created via the private method '_delete_tag'.

Arguments:

(+) $index1, $index2: The start index and the end index of the tag range to be removed

=cut

##############################################################################
# Alle <PAR_STYLES>-Tags die im Bereich von index1 bis index2 liegen loechen #
##############################################################################
sub delete_par_tags_in_area
{
my $self = shift;
my $index1 = shift;
my $index2 = shift;
my $text_widget = $self->docform_doctext();
$index1 = $text_widget->index("$index1"); #Transform into line.char 
$index2 = $text_widget->index("$index2"); #Transform into line.char 
my $col1;
my $col2;
my $line1;
my $line2;
if ($index1 =~ /^(\d+).(\d+)$/o )
	{
	$line1 = $1;
	$col1= $2;
	}
if ($index2 =~ /^(\d+).(\d+)$/o )
	{
	$line2 = $1;
	$col2= $2;
	}
my %par_tags; #Liste als Hash (Um doppelte Eintragungen zu verhindern), die alle gefundenen <PAR_STYLES>-Tags im Bereich aufnehmen soll
#Durch alle Zeilen des Bereiches gehen
for (my $l=$line1; $l<=$line2; $l++) 
	{
	# Anfangs- und Endspalte der aktuellen Zeile ermitteln
	my $startcol;
 	my $endcol; 
	if ($l==$line1) {$startcol=$col1} else {$startcol=0};
	if ($l==$line2) {$endcol=$col2} else {$endcol=$text_widget->index("$l.end")}; 
	# Aktuelle Zeile Zeichenweise durchgehen und PAR_STYLES suchen
	for (my $c=$startcol; $c<=$endcol; $c++)
   	{
		# Alle Tags an dieser Position auslesen
		my @tags = $text_widget->tagNames("$l.$c");
		# Durch alle gefundenen Tags gehen
		foreach my $tag (@tags)
			{
			# Wenn ein Tag gefunden wird
			if ($tag =~ /^<PAR_STYLES>/o)
				{
				# In Ergebnisliste (Hash) eintragen
				$par_tags{$tag}=1;
				}
			}
		}
	}
# Gefundene <PAR_STYLES>-Tags loeschen
foreach my $key (keys %par_tags)
	{
	$self->_delete_tag($key,'-1','-1');
	}
}

=head3 delete_env_tags_in_area

$object->delete_env_tags_in_area($index1, $index2);

Removes all <ENV_STYLES> tag from the textwidget and the $self->{'tags'} branch. All defined
arguments are also deleted in the $self->{'args'} branch.

The tags and arguments are created via the private method '_delete_tag'.

Arguments:

(+) $index1, $index2: The start index and the end index of the tag range to be removed

=cut

##############################################################################
# Alle <ENV_SYTLES>-Tags die im Bereich von index1 bis index2 liegen loechen #
##############################################################################
sub delete_env_tags_in_area
{
my $self = shift;
my $index1 = shift; 
my $index2 = shift;
my $text_widget = $self->docform_doctext();
$index1 = $text_widget->index($index1); #Transform into line.char 
$index2 = $text_widget->index($index2); #Transform into line.char 
# transforming $index1 and $index2 into lines an collumns
my $col1;
my $col2;
my $line1;
my $line2;
if ($index1 =~ /^(\d+).(\d+)$/o )
	{
	$line1 = $1;
	$col1= $2;
	}
if ($index2 =~ /^(\d+).(\d+)$/o )
	{
	$line2 = $1;
	$col2= $2;
	}
my %env_tags; #Liste als Hash (Um doppelte Eintragungen zu verhindern), die alle gefundenen <PAR_STYLES>-Tags im Bereich aufnehmen soll
#Durch alle Zeilen des Bereiches gehen
for (my $l=$line1; $l<=$line2; $l++) 
	{
	# Anfangs- und Endspalte der aktuellen Zeile ermitteln
	my $startcol;
 	my $endcol; 
	if ($l==$line1) {$startcol=$col1} else {$startcol=0};
	if ($l==$line2) {$endcol=$col2} else {$endcol=$text_widget->index("$l.end")}; 
	# Aktuelle Zeile Zeichenweise durchgehen und ENV_STYLES suchen
	for (my $c=$startcol; $c<=$endcol; $c++)
   	{
		# Alle Tags an dieser Position auslesen
		my @tags = $text_widget->tagNames("$l.$c");
		# Durch alle gefundenen Tags gehen
		foreach my $tag (@tags)
			{
			# Wenn ein Tag gefunden wird
			if ($tag =~ /^<ENV_STYLES>/o)
				{
				# In Ergebnisliste (Hash) eintragen
				$env_tags{$tag}=1;
				}
			}
		}
	}

# Gefundene <ENV_STYLES>-Tags loeschen
foreach my $key (keys %env_tags)
	{
	$self->_delete_tag($key,$index1,$index2);
	}
}

=head3 delete_char_tags_in_area

$object->delete_char_tags_in_area($index1, $index2);

Removes all <CHAR_STYLES> tag from the textwidget and the $self->{'tags'} branch. All defined
arguments are also deleted in the $self->{'args'} branch.

The tags and arguments are created via the private method '_delete_tag'.

Arguments:

(+) $index1, $index2: The start index and the end index of the tag range to be removed

=cut


###############################################################################
# Alle <CHAR_STYLES>-Tags die im Bereich von index1 bis index2 liegen loechen #
###############################################################################
sub delete_char_tags_in_area
{
my $self = shift;
my $index1 = shift;
my $index2 = shift;
my $text_widget = $self->docform_doctext();
$index1 = $text_widget->index($index1); #Transform into line.char 
$index2 = $text_widget->index($index2); #Transform into line.char 
# transforming $index1 and $index2 into lines an collumns
my $col1;
my $col2;
my $line1;
my $line2;
if ($index1 =~ /^(\d+).(\d+)$/o )
	{
	$line1 = $1;
	$col1= $2;
	}
if ($index2 =~ /^(\d+).(\d+)$/o )
	{
	$line2 = $1;
	$col2= $2;
	}
my %char_tags; #Liste als Hash (Um doppelte Eintragungen zu verhindern), die alle gefundenen <PAR_STYLES>-Tags im Bereich aufnehmen soll
#Durch alle Zeilen des Bereiches gehen
for (my $l=$line1; $l<=$line2; $l++) 
	{
	# Anfangs- und Endspalte der aktuellen Zeile ermitteln
	my $startcol;
 	my $endcol; 
	if ($l==$line1) {$startcol=$col1} else {$startcol=0};
	if ($l==$line2) {$endcol=$col2} else {$endcol=$text_widget->index("$l.end")}; 
	# Aktuelle Zeile Zeichenweise durchgehen und CHAR_STYLES suchen
	for (my $c=$startcol; $c<=$endcol; $c++)
   	{
		# Alle Tags an dieser Position auslesen
		my @tags = $text_widget->tagNames("$l.$c");
		# Durch alle gefundenen Tags gehen
		foreach my $tag (@tags)
			{
			# Wenn ein Tag gefunden wird
			if ($tag =~ /^<CHAR_STYLES>/o)
				{
				# In Ergebnisliste (Hash) eintragen
				$char_tags{$tag}=1;
				}
			}
		}
	}

# Gefundene <CHAR_STYLES>-Tags loeschen
foreach my $key (keys %char_tags)
	{
	$self->_delete_tag($key,$index1,$index2);
	}
}

=head3 toggle_tagdoctor_tags()

$object->toggle_tagdoctor_tags();

Switches the tag-doctor tags on/off in the whole document

=cut

#####################################################
### Build/Remove the TagDoc objects in all lines  ###
#####################################################
sub toggle_tagdoctor_tags
{
my $self = shift;
my $doctext = $self->docform()->doctext();
my @tagnames = $doctext->tagNames();
#Set default value for the first call
my $end_line = $doctext->index('end');
$end_line =~ s/\.(.*)$//o;
if ($self->docform()->tag_doctor_on() == 1)
	{
	#Preventing multiple calls -> could corrupt the document
	$self->docform()->{'ui'}->{'tagdoctor_button'}->configure(-state=>'disabled');
	#Initialise the progress bar
	$self->docform()->_progress_init('Sending home the TagDoctor');
	#Delete all 'TAG_DOC'-tags
	my @ranges = $doctext->tagRanges('TAG_DOC');
	my $number = $#ranges;
	my $counter = 0;
	my $last_10_percent_block=0;
	#Iterate over the tag-ranges
	while ($#ranges>=0)
		{
		$counter++;
		#Updating progress bar%
		my $now = int(int($counter/$number*100)/10);
		if ($now > $last_10_percent_block)
			{
			$self->docform()->_progress_update('He is going ...',int($counter/$number*100));
			$last_10_percent_block = $now;
			}
		#Remove the ranges
		my $end_index = pop(@ranges);
		my $start_index= pop(@ranges); 
		$doctext->delete($start_index, $end_index);
		}
	#Setting the property
	$self->docform()->tag_doctor_on( 0 );
	#Terminating the progress bar
	$self->docform()->_progress_finish();
	#Preventing multiple calls -> could corrupt the document
	$self->docform()->{'ui'}->{'tagdoctor_button'}->configure(-state=>'active');
	}
else
	{
	#Preventing multiple calls -> could corrupt the document
	$self->docform()->{'ui'}->{'tagdoctor_button'}->configure(-state=>'disabled');
	#Initialise the progress bar
	$self->docform()->_progress_init('Calling for the TagDoctor');
	my @tagnames = $doctext->tagNames();
	#Ensuring the correct order of the different tag types
	@tagnames = $self->sort_tags_by_stylegroup($doctext);
	my $number = $#tagnames;
	my $counter = 0;
	my $last_10_percent_block=0;
	#Iterate over the tagnames
	while ($#tagnames>=0)
		{
		my $tagname = pop(@tagnames);
		$counter++;
		#Updating progress bar
		my $now = int(int($counter/$number*100)/10);
		if ($now > $last_10_percent_block)
			{
			$self->docform()->_progress_update('He is coming ...',int($counter/$number*100));
			$last_10_percent_block = $now;
			}
		#Insert the TagDoc-objects
		if ($tagname =~ /^<PAR_STYLES>/ or $tagname =~ /^<ENV_STYLES>/ or $tagname =~ /^<CHAR_STYLES>/)
			{
			$self->insert_tag_doctor_objects($tagname, undef, undef);
			}
		}
	#Setting the property
	$self->docform()->tag_doctor_on( 1 );
	#Terminating the progress bar
	$self->docform()->_progress_finish();
	#Preventing multiple calls -> could corrupt the document
	$self->docform()->{'ui'}->{'tagdoctor_button'}->configure(-state=>'active');
	}
}


=head3 insert_tag_doctor_objects

$object->insert_tag_doctor_objects($tagname, $stylegroup, $style);

Frame the tag with 2 TagDoc objects.

Parameters:

(+) $tagname: The name of the tag to border

(+) $stylegroup: The stylegroup name  of the tag (if set to undef -> extracted from the tagname)

(+) $style: The style name  of the tag  (if set to undef -> extracted from the tagname)

=cut

#############################################
### Build the TagDoc objects in one line  ###
#############################################
sub insert_tag_doctor_objects
{
my $self = shift;
my $tagname = shift;
my $stylegroup = shift;
my $style = shift;

#Extract the stylegroup and style from the tagname, if not defined
if (not defined $stylegroup or not defined $style)
	{
	if ($tagname =~ /^<(.+?)><(.+?)><(.*?)>:\d+$/o)
		{
		$stylegroup = $1;
		$style = $2;
		}
	}

my $config = $main::config;
my $doctext = $self->docform()->doctext();

my $colors={};
$colors->{'PAR_STYLES'} = $config->{'UI'}->{'par_styles_background'};
$colors->{'ENV_STYLES'} = $config->{'UI'}->{'env_styles_background'};
$colors->{'CHAR_STYLES'} = $config->{'UI'}->{'char_styles_background'};

#my %offset_in_line; #A hash that holds the actual TagDoc-object-caused offset per line
#my $actual_pair_offset; #The offset caused by the first of the 2 TagDoc objects that belong to a tags range

my @ranges = $doctext->tagRanges($tagname);
while ($#ranges>=0)
	{
	my $end_index = pop(@ranges);
	my $start_index= pop(@ranges); 
	my $start_line = Utils::get_index_info($start_index);
	my $end_line = Utils::get_index_info($end_index);
	my $offset=0; #Offset for the end-tag if both are in the same line
	#Initialise the offset-hash, if there has not been a range inside of this line
	#$offset_in_line{$start_line} = 0 if not exists $offset_in_line{$start_line} or not defined $offset_in_line{$start_line};
	#$offset_in_line{$end_line} = 0 if not exists $offset_in_line{$end_line} or not defined $offset_in_line{$end_line};
	$offset = 1 if $start_line == $end_line;
	#Create the start-TagDoc object
	my $starttag = $doctext->TagDoc(-text=>"<".$style.">", -background=>$colors->{$stylegroup}, -text_widget=>$doctext, -associated_tag=>$tagname, -type=>'start', -tagman=>$self);
	#Create the start-TagDoc object
	my $endtag = $doctext->TagDoc(-text=>"</".$style.">", -background=>$colors->{$stylegroup}, -text_widget=>$doctext, -associated_tag=>$tagname, -type=>'end', -tagman=>$self);
	#Embed the TagDoc-object into the text widget
	my $end_window = $doctext->windowCreate($end_index, -window=>$endtag, -stretch=>0, -align=>'center');
	#Marking the TagDoc for removal
	$doctext->tagAdd('TAG_DOC', $end_index);
	#Embed the TagDoc-object into the text widget
	my $start_window = $doctext->windowCreate($start_index, -window=>$starttag, -stretch=>0, -align=>'center');
	#Marking the TagDoc for removal
	$doctext->tagAdd('TAG_DOC', $start_index);
	#If the Labels are removed,
	#the tag has to be extended over the 2 labels, it will shrink automatically to the old size
	$doctext->tagAdd($tagname, $start_index, "$end_index + ".($offset+1)." chars");
	#Increase the offset-counters for the lines where the TagDoc objects have been created
	#$offset_in_line{$start_line}++ ;
	#$offset_in_line{$end_line}++ ;
	}
}

=head3 remove_tag_doctor_objects

$object->remove_tag_doctor_tags($tagname, $stylegroup, $style);

Remove the TagDoc-Objects from aspecific tag

Parameters:

(+) $tagname: The name of the tag to border

(+) $stylegroup: The stylegroup name  of the tag (if set to undef -> extracted from the tagname)

(+) $style: The style name  of the tag  (if set to undef -> extracted from the tagname)

=cut

##############################################
### Remove the TagDoc objects in one line  ###
##############################################
sub remove_tag_doctor_objects
{
my $self = shift;
my $tagname = shift;
my $style = shift;
my $stylegroup = shift;
my $config = $main::config;
my $doctext = $self->docform()->doctext();

#Extract the stylegroup and style from the tagname, if not defined
if (not defined $stylegroup or not defined $style)
	{
	if ($tagname =~ /^<(.+?)><(.+?)><(.*?)>:\d+$/o)
		{
		$stylegroup = $1;
		$style = $2;
		}
	}

my $found_one = 0; #Flag
my @ranges = $doctext->tagRanges('TAG_DOC');
while ($#ranges>=0)
	{
	my $start_index= shift(@ranges); 
	my $end_index = shift(@ranges);
	my $tag_doc = $doctext->windowCget($start_index, '-window');
	if (defined $tag_doc)
		{
		my $assoc_tag = $tag_doc->associated_tag();
		if ($assoc_tag eq $tagname)
			{
			$found_one = 1; #Tag is found, do it again

			#HERE SHOULD COME CODE TO BREAK UP CIRCLE REFERENECES
			$tag_doc->do_destroy();

			#DELETE THE TAGDOC FROM THE TEXT WIDGET
			$doctext->delete($start_index, $end_index);

			}
		}
	if ($found_one  == 1)
		{
		@ranges = $doctext->tagRanges('TAG_DOC');
		$found_one = 0;
		}
	}
}

=head3 get_num_list

$object->get_num_list($stylegroup, $style);

Return a reference to the num_list that is correlated with the given style, or undef if there is none.

Parameters:

(+) $stylegroup: The stylegroup name 

(+) $style: The style name 

=cut

######################################
### Get a Ref to a NumList object  ###
######################################
sub get_num_list
{
my $self = shift;
my $stylegroup = shift;
my $style = shift;
my $internal = $main::internal;
my $numlist= undef;
my $name = $internal->{$stylegroup}->{$style}->{'connect_to_doc_numlist'};
if (defined $name)
	{
	$numlist = $self->{'doc_numlists'}->{$name};
	}
return $numlist;
}


=head2 Private methods:

This methods are for the objects internal use only, do not call them from outside the object.

=cut

###################
# Private methods #
###################

=head3 _tag_counter

$value = $object->_tag_counter(); #gets the actual value of the tag counter 

$object ->_tag_counter($value); #sets the actual value of the tag counter (DO NOT USE, use _inc_tag_counter method for increasing it)

=cut

### _tag_counter ###
sub _tag_counter
{
my $self=shift;
@_ ? $self->{'tag_counter'} = shift : return $self->{'tag_counter'};
}

=head3 _inc_tag_counter

$new_value = $object ->_inc_tag_counter(); #increases the actual value of the tag counter by 1

=cut

##################################
### Increase tag counter by 1  ###
##################################
sub _inc_tag_counter
{
my $self=shift;
my $new_value = $self->_tag_counter()+1; #increase it
$self->_tag_counter($new_value); #set it
return $new_value; #return it
}

=head3 _clear_tags

$object->_clear_tags();

Clears all stored data in the $self->{'tags'} branch.

=cut

######################
# Tagsliste loeschen #
######################
sub _clear_tags
{
my $self = shift;

$self->{'tags'}={};
}

=head3 _build_list_items

$object->_build_list_items($list_tagname, $index1, $index2);

This private method creates all needed NumLabel objects for building up
a numbered or a bullet list. After that the corresponding NumList object
is renumbered.

Unfortionatly it cannot handle nested lists yet.

Arguments:

(+) $list_tagname: The tagname of the list to build up the items for

(+) $index1, $index2: The start index and the end index of the tag range, if both are set
to -1, it means that the whole range of the tag is to be (re)build.

=cut

###################################
# Alle Items einer Liste aufbauen #
###################################
sub _build_list_items
{
my $self = shift;
my $list_tagname = shift;
my $index1=shift; # ist -1 wenn ganzer Tagbereich gemeint ist
my $index2=shift; # ist -1 wenn ganzer Tagbereich gemeint ist
my $text_widget = $self->docform_doctext();

#Tag in Einzelteile zerlegen
my $style;
my $stylegroup;
my $number;
if ($list_tagname =~ /^<(.+?)><(.+?)><(.*?)>:(.+?)$/o)
	{
	$stylegroup = $1;
	$style = $2;
	$number = $4;
	}

# ggf. Bereiche aus TagMan holen
if ($index1 == -1 and $index2== -1)
	{
	my @range = $text_widget->tagRanges($list_tagname);
	$index1 = shift(@range);
	$index2 = shift(@range);
	}
#In Zeilen + Spalten umwandeln
my $startline = $index1;
$startline =~ s/\.(.*)$//go;
my $endline = $index2;
$endline =~ s/\.(.*)$//go;
#A hash that keeps a flag for each line in the range wether ther is already an listitem or not.
my %linehash;
#A varaible that stores the highest tag-number of the already used LIST_NUM-tags
my $highest_tagnumber = 0;
# Referenz auf NumList
my $num_list = $self->{'env_numlists'}->{$list_tagname}; # Referenz auf NummernlabelListen-Objekt uebergeben
#Iterate over all NumLabels that are managed by this numlist and mark the row in the hash %linehash
foreach my $numlabel (@{$num_list->list()})
	{
	$linehash{$numlabel->get_current_line()} = 1;
	if ($numlabel->window_tag() =~ /:(.*)$/)
		{
		$highest_tagnumber = $1 if $1 > $highest_tagnumber;
		}
	}

#Iterate over the whole range of the list
for (my $line=$startline; $line<=$endline; $line++)
	{
   #If there isnt already an NumLabel in the line
	if (not defined $linehash{$line})
   	{
		#Eigenschaften fuer die NumLabel-Objekte aus $main::nternal-Hash holen
		my $level = '1';
		my $window_tag = "<LIST_NUM><$style><$number>:".($highest_tagnumber+$line-$startline+1); # Namen des Markierungstags fuer das NumLabel
		my $internal_branch = $main::internal->{$stylegroup}->{$style};
		my $num_font=$internal_branch->{'num_font'};
		my $num_borderwidth=$internal_branch->{'num_borderwidth'};
		my $num_background=$internal_branch->{'num_background'};
		my $num_foreground=$internal_branch->{'num_foreground'};
		my $num_relief=$internal_branch->{'num_relief'};
		my $num_pre_number_text = $internal_branch->{'subitems_pre_number_text'};
		my $num_post_number_text = $internal_branch->{'subitems_post_number_text'};
		my $num_show_number = $internal_branch->{'subitems_show_number'};
		my $num_show_bullet = $internal_branch->{'subitems_show_bullet'};
		#Nummernlabel erzeugen
		my $number_label = $text_widget->NumLabel(-numlist=>$num_list, -level=>$level, -pre_number_text=>$num_pre_number_text, -post_number_text=>$num_post_number_text, -text_widget=>$text_widget, -associated_tag=>$window_tag, -relief=>$num_relief, -font=>$num_font,-background=>$num_background,-foreground=>$num_foreground,-borderwidth=>$num_borderwidth,-show_number=>$num_show_number, -show_bullet=>$num_show_bullet);
		#Fenster in Textwidget erzeugen und Nummernlabel dort einbetten
		my $offset=0;#Position of the Numlabel (normaly $line.0)
		#But if this is the first line and the Tag-Doctor is turned on, it has to be the second object in the first line
		if (($line==$startline) and ($self->docform()->{'_tag_doctor_on'} eq 'TAG-DOCTOR:ON'))
			{
			#Test if its really there
			my @tags=$self->docform()->doctext()->tagNames("$line.0");
			foreach my $tag (@tags)
				{
				$offset = 1 if $tag eq 'TAG_DOC';
				}
			}
		my $absnummerwindow = $text_widget->windowCreate("$line.$offset", -window=>$number_label);
		#Tag um Nummernlabel herum erzeugen
		$text_widget->tag("add",$window_tag,"$line.$offset","$line.$offset + 1 chars"); #Eingebettets Nummernlabel mit einem <Num>-Tag markieren um es von den <PAR_STYLES>-Tags unterscheiden zu koennen (Hat aber die selbe Nummer wie <PAR_STYLES>-Label!!!)
		#Markierungstag-Name an NumLabel-Objekt uebergeben
		$number_label->window_tag( $window_tag );
		#NumLabel-Objekt fertig initialisieren
		$number_label->on_create(0);
		}
	}
#Span ENV-Tag over the full range, including the NumLabels
$text_widget->tagRemove($list_tagname,"$startline.0","$endline.end");
$text_widget->tagAdd($list_tagname,"$startline.0","$endline.end");

#Neu durchnummerieren
$num_list->renumber();
}

=head3 _remove_list_items

$object->_remove_list_items($list_tagname, $index1, $index2);

This private method removes all existing NumLabel objects for building up
a numbered or a bullet list. 

Unfortionatly it cannot handle nested lists yet.

Arguments:

(+) $list_tagname: The tagname of the list to build up the items for

(+) $index1, $index2: The start index and the end index of the tag range, if both are set
to -1, it means that the whole range of the tag is to be (re)build.

=cut

####################################
# Alle Items einer Liste entfernen #
####################################
sub _remove_list_items
{
my $self = shift;
my $list_tagname = shift;
my $index1 = shift; #Ist -1, wenn ganzes Tag gemeint ist
my $index2 = shift; #Ist -1, wenn ganzes Tag gemeint ist
my $text_widget = $self->docform_doctext();
my $style;
my $stylegroup;
my $number;

#If this ENV_STYLE tag is no numbered or buller list ->exit
if (not exists $self->{'env_numlists'}->{$list_tagname} or not defined $self->{'env_numlists'}->{$list_tagname}) {goto EXIT}; 

#Tagnamen zerlegen
if ($list_tagname =~ /^<(.+?)><(.+?)><(.*?)>:(.+?)$/o)
	{
	$stylegroup = $1;
	$style = $2;
	$number = $4;
	}

#ggf. Bereiche aus TagMan-Objekt holen
if (($index1 ==-1)&&($index2==-1))
	{
	$self->read_from_widget($text_widget);
	$index1 = $self->{'tags'}->{$list_tagname}->{0}->{'start_index'};
	$index2 = $self->{'tags'}->{$list_tagname}->{0}->{'end_index'};
	}
#Bereich in Zeilen und Spalten zerlegen
my $startline = $index1;
$startline =~ s/\.(.*)$//go;
my $endline = $index2;
$endline =~ s/\.(.*)$//go;
#Referenz auf NumList-Objekt holen
my $num_list = $self->{'env_numlists'}->{$list_tagname}; # Referenz auf NummernlabelListen-Objekt uebergeben
$num_list->clear(); #Clear the whole numlist
# Zeilenweise ueber Bereich iterieren
for (my $line=$startline; $line<=$endline; $line++)
	{
	#iterate over all lines of the list
	my @tags;
	my $ende = $text_widget->index("$line.end");
	my $columns;
	if ($ende =~ /^(.+?)\.(.+?)$/o) {$columns=$2};
	my %tags;
	# iterate over the line and add found tags to array
	for (my $y=0; $y<=$columns; $y++)
		{
		my @some_tags = $text_widget->tagNames("$line.$y");
		foreach my $some_tag (@some_tags)
			{
			$tags{$some_tag}=1;
			}
		}
	# delete duplicate entries
	while ((my $key,my $value) = each %tags)
		{
		push(@tags,$key);
		}
	#Alle gefundenen Tags zerlegen
	foreach my $tag (@tags)
		{
		#If its an LIST_NUM-Tag
		if ($tag =~ /^<LIST_NUM>/o)
			{
			#Wenn es ein NummernLabel einer Liste ist
			my @range = $text_widget->tagRanges($tag);
			#NumLabel lschen
			$text_widget->delete(@range);
			}
		}
	}
EXIT:
}

=head3 _build_display_options_hash

%options = $object->_build_display_options_hash($stylegroup, $style);

Retrieves the options for displaying from the hash $main::internal and caches
them in a branch $main::internal->{STYLEGROUP_NAME}->{STYLE_NAME}->{DISPLAY_OPTIONS}

The functions returns the options for configuring the tag (CHAR_STYLES, PAR_STYLES, ENV_STYLES) or the widget (COMMANDS, EMB_OBJ).

Parameters:

(+) $stylegroup: The name of the stylegroup

(+) $style: The name of the style

=cut

##########################################################################
# Builds a hash for the display-options for the given tag/command/embobj #
##########################################################################
sub _build_display_options_hash
{
my $self = shift; #Zeiger auf Objekt
my $stylegroup = shift;
my $style = shift;
my %options;
my $internal_branch = $main::internal->{$stylegroup}->{$style};
#If the hash has been already build up
if ( exists $internal_branch->{'DISPLAY_OPTIONS'} and defined $internal_branch->{'DISPLAY_OPTIONS'} )
	{
	#Get the cahched one
	%options = %{$internal_branch->{'DISPLAY_OPTIONS'}};
	}
else
	{
	#Get the names of all possible display-options
	my @display_options = @{$self->{'display_options'}};
	#Iterate through them
	for (my $i=0;$i<= $#display_options;$i++) #Durch alle definierten Optionen zur Tagdarstellung gehen
		{
		my $option=$display_options[$i]; #Eine Option auslesen
		#If this option was set in 'edoc.int'
		if (exists($internal_branch->{$option})) #Wenn diese in INTERNAL-Struktur definiert ist
			{
			#Add it to the hash
			$options{"-$option"}=$internal_branch->{$option}; #In Options-Hash eintragen
			}
		}
	#Cache it
	$internal_branch->{'DISPLAY_OPTIONS'} = \%options;
	}
return %options;
}


=head3 _before_add_tag

$tagnem = $object->_before_add_tag($stylegroup, $style, $index1, $index2, $do_not_sort_par_numlabel);

This is the main function where tags in the text widget are created. This ensures the correct
registration of the tags. 

The method '_after_add_tag' is called after the calling method of '_before_add_tag' has completed all things it does
with this newly created tag to finnish it up. 

For PAR_STYLE tags the NumLabels are created and for ENV_STYLE tags that represent lists
are the item buildup methods called.

The method returns the full tagname of the newly created tag. 

Arguments:

(+) $stylegroup: The style group the tag belongs to. (PAR_STYLES, ENV_STYLES, CHAR_STYLES, COMMANDS, EMB_OBJ)

(+) $stlye: The name of the style to add

(+) $index1, $index2: The start index and the end index of the tag range

(+) $do_not_sort_par_numlabel: If this flag is set to 1 the possibly automatically created NumLabel should not be sorted and renumbered
to save time on document opening when many tags are created and one renumbering at the end is enough. This is why $self->renumber()
has to be called at the end of those 'file_open' functions. (Is only working with the stylegroup 'PAR_STYLES')

=cut

####################
# Ein Tag erzeugen #
####################
sub _before_add_tag
{
my $self = shift; #Zeiger auf sich selbst
my $stylegroup = shift; #Formatgruppe
my $style = shift; #Absatzformat-Name
my $index1 = shift; #Begin-Index
my $index2 = shift; #End-Index
my $do_not_sort_par_numlabel = shift; #Fuer PAR_STYLES, wenn NumLabels nicht sortiert werden sollen 
my $text_widget = $self->docform_doctext();
my $counter = $self->_tag_counter();
my $tag_name = "<$stylegroup><$style><>:$counter";
my $internal_branch = $main::internal->{$stylegroup}->{$style};

#Calculating linenumbers
$index1 = $text_widget->index($index1);
$index2 = $text_widget->index($index2);
(my $startline, my $startcol) = $index1 =~ /^(\d+?)\.(\d+?)$/o;
(my $endline, my $endcol) = $index2 =~ /^(\d+?)\.(\d+?)$/o;

# Tag  darberlegen
$text_widget->tagAdd($tag_name,$index1,$index2);

# Setting the optionf of the tag for all styles that ere not COMMANDS or EMBOBJ
if ($stylegroup =~ /_STYLES$/) #If its PAR_STYLES, ENV_STYLES or CHAR_STYLES
	{
	my %options = $self->_build_display_options_hash($stylegroup, $style); #Get display options
	$text_widget->tagConfigure($tag_name, %options);
	}

# Tagcounter erhoehen
$self->_inc_tag_counter();

EXIT:
#Tagnamen zurueckliefern
return $tag_name;
}

=head3 _after_add_tag

$object->_after_add_tag($tagname, $stylegroup, $style, $allow_add_tagdoc);

This is called after all additional stuff with the tag (adding of NumLabels, PreText, PostText, etc...)
is done and the definite range of characters is tagged with the tag.

Arguments:

(+) $tagname: The name of the tag where it happened

(+) $stylegroup: The stylegroup of the tag, if undef -> extracted from the tagname

(+) $style: The style of the tag, if undef -> extracted from the tagname

(+) $allow_add_tagdoc: If set to 0 or undef, do not add TagDoc-objects (disturbing at loading/paste-time)

=cut

##################################################
# Code that should finnish up the tag insertion  #
##################################################
sub _after_add_tag
{
my $self = shift; #Zeiger auf das TagmanObjekt
my $tag_name = shift; #Name des aufrufenden Listen-Tags
my $stylegroup = shift;
my $style = shift;
my $allow_add_tagdoc = shift;

my $text_widget = $self->docform_doctext();

$allow_add_tagdoc = 0 if not defined $allow_add_tagdoc;

#Get stylegroup and style from tagname, if not defined in the paramters
if (not defined $stylegroup or not defined $style)
	{
	if ($tag_name =~ /^<(.+?)><(.+?)><(.*?)>:\d+$/o)
		{
		$stylegroup = $1;
		$style = $2;
		}
	}

# Eintragung in Argumente-Struktur durchfuehren
my $arg_number = $self->_add_arguments($tag_name,$stylegroup,$style);

# Rechtes Mausklickereignis binden (erst jetzt, da es zuvor keine Argumente gab)
# und nur wenn es Argumente zu diesem Stil gibt.
#Ebenso wird dann der Mauszeiger adaptiert
if ($arg_number>0)
	{
	$text_widget->tagBind($tag_name,'<Button-3>' => sub{$self->_do_tag_click($tag_name)});
	#Wenn zu einem Tag ein Argument eingefuehrt wird -> Hyperlink-Cursor anzeigen
	$self->docform_doctext()->tagBind($tag_name,'<Any-Enter>' , sub{shift->configure(-cursor=>'cross')});
	$self->docform_doctext()->tagBind($tag_name,'<Any-Leave>' , sub{shift->configure(-cursor=>'xterm')});
	}


#Refresh the 'TAG_DOC'-objects if the TagDoc feature is turned on 
if ($self->docform()->tag_doctor_on()==1 and $allow_add_tagdoc == 1)
	{
	$self->insert_tag_doctor_objects($tag_name, $stylegroup, $style);
	}
}


=head3 _return_pressed

$object->_return_pressed($tagname);

This is a callback that is automatically bind to the hitting of the 'Enter' key
inside of a numbered list or a bullet list. It calls the _build_list_items method
to ensure to correct renumbering or setting of bullets.

Arguments:

(+) $tagname: The name of the tag where it happened

=cut

############################################
# Callback for Return-Key inside of a list #
############################################
sub _return_pressed
{
my $self = shift; #Zeiger auf das TagmanObjekt
my $tag_name = shift; #Name des aufrufenden Listen-Tags
#Items neu aufbauen und durchnummerieren
$self->_build_list_items($tag_name,-1,-1);
}

=head3 _delete_tag

$object->_delete_tag($tagname, $index1, $index2);

Removes the tag with the given tagname from the area mentioned by $index1 and $index2.

If the whole tag was removed, the arguments are deleted too from the $self->{'args'} branch using
the method _delete_arguments.

Arguments:

(+) $tagname: The name of the tag to remove

(+) $index1, $index2: The start index and the end index of the removal range (If both are set to -1, the whole tagrange is removed.
For <PAR_SYTLES> the indexes have to be set to -1 because they can only be removed as a whole.

=cut

###################
# Loescht ein Tag #
###################
sub _delete_tag
{
my $self = shift;
my $tagname = shift;
my $index1 = shift; #ist '-1' bei PAR_STYLES, da diese nur als gesamtheit geloescht werden sollen
my $index2 = shift; #ist '-1' bei PAR_STYLES, da diese nur als gesamtheit geloescht werden sollen
my $text_widget = $self->docform_doctext();
my $name=$tagname; #Tagnamen merken

#Calculating linenumbers
my $startline = $index1;
my $endline = $index2;
if ($startline == -1 and $endline==-1) 
	{
	my @range = $text_widget->tagRanges($tagname);
	$startline=shift(@range);
	$endline=shift(@range);
	}
$startline =~ s/\.(.*)$//go;
$endline =~ s/\.(.*)$//go;

# bei PAR_STYLES muss auf NUM-Tag geprueft werden, dort sind $index1 und $index2 egal,
# da PAR_STYLES immer nur 1 mal vorkommen und ueber eine ganze Zeile gehen und nur
# als Gesamtheit geloescht werden koennen
if ($tagname =~ /^<PAR_STYLES>/o)
	{
	# Wenn es ein dazugehoerendes <NUM>-Tag gibt ->loeschen
	my @tags = $text_widget->tagNames();
	foreach my $numname (@tags)
		{
		if ($numname =~ /^<NUM>/o)
			{
			my $search=$name;
			$search =~ s/^<PAR_STYLES>/<NUM>/o ;
			if ($numname eq $search)
				{
				my @range = $text_widget->tagRanges($numname);
				if ($#range>=0) #Nur Loeschen, wenn Nummernlabel noch existiert, sonst kommt Fehler
					{
					$text_widget->delete($text_widget->tagRanges($numname));
					}
				}
			}
		}
	# Ganzes Tag loeschen
	$text_widget->tagDelete($tagname);
	# Wenn es gespeicherte Argumente zum Tag gibt ->loeschen
	$self->_delete_arguments($tagname);
	}

elsif ($tagname=~ /^<ENV_STYLES>/o) 
	{
	# Remove listitems, if there
	$self->_remove_list_items($tagname,$index1,$index2);
	# Tag aus Bereich wegnehmen, wenn ein Bereich definiert ist
	$text_widget->tagRemove($tagname,$index1,$index2);
	# Wenn Tag keinen Bereich mehr hat
	my @range = $text_widget->tagRanges($tagname);
	if (($#range>0)==0)
		{
		# Ganzes Tag loeschen
		$text_widget->tagDelete($tagname);
		# Wenn es gespeicherte Argumente zum Tag gibt ->loeschen
		$self->_delete_arguments($tagname);
		# Bei ENV_STYLE-Tags ggf. auch Nummernliste bei nummerierten Listen und Bulletlisten loeschen
		if ((exists $self->{'env_numlists'}->{$tagname})&&(defined $self->{'env_numlists'}->{$tagname}))
			{
			$self->{'env_numlists'}->{$tagname} = undef;
			}
		}
	}
else
	{
	# Tag aus Bereich wegnehmen, wenn ein Bereich definiert ist
	$text_widget->tagRemove($tagname,$index1,$index2);
	# Wenn Tag keinen Bereich mehr hat
	my @range = $text_widget->tagRanges($tagname);
	if (($#range>0)==0)
		{
		# Ganzes Tag loeschen
		$text_widget->tagDelete($tagname);
		# Wenn es gespeicherte Argumente zum Tag gibt ->loeschen
		$self->_delete_arguments($tagname);
		}
	}

#Refresh the 'TAG_DOC'-tags if it is turned on
if ($self->docform()->tag_doctor_on()==1)
	{
	$self->remove_tag_doctor_objects($tagname, undef, undef);
	}
}

=head3 _delete_arguments

$object->_delete_arguments($tagname);

Deletes the stored arguments in the $self->{'args'} branch for the given tag.

Arguments:

(+) $tagname: The name of the tag which arguments should be deleted

=cut

##########################################
# Speicherplatz fuer Argumente freigeben #
##########################################
sub _delete_arguments
{
my $self = shift;
my $tag_name = shift;
if (exists $self->{'args'}->{$tag_name})
	{
	delete $self->{'args'}->{$tag_name};
	}
}

=head3 _add_arguments

$arg_number = $object->_add_arguments($tagname, $stylegroup, $style);

Initalises the arguments in the $self->{'args'} branch for the given tagname.
The stylegroup and stylename have to be known to retrieve the argument information 
from the data out of the 'edoc.int' file.

The method returns the number of initialised arguments.

Arguments:

(+) $tagname: The name of the tag which arguments should be initialised

(+) $stylegroup: The style group the tag belongs to

(+) $style: The style the tag belongs to

=cut

###############################################
# Speicherplatz fuer Argumente initialisieren #
###############################################
sub _add_arguments
{
my $self = shift;
my $tagname = shift;
my $style_group = shift;
my $style = shift;

my $internal_branch = $main::internal->{$style_group}->{$style};

my $arg_number = $internal_branch->{'arg_number'};

$arg_number = 0 if not defined $arg_number;

if ($style_group eq 'EMBOBJ')
	{
	my $object = $self->{'objects'}->{$tagname};
	$object->{'args'}->{'arg_number'}=$arg_number;
	}
elsif ($style_group eq 'DOC_STYLES')
	{
	$self->{'doc_args'}->{'arg_number'} = $arg_number;
	}
else
	{
	$self->{'args'}->{$tagname}->{'arg_number'}=$arg_number;
	}

if ($arg_number > 0)
	{
	for (my $i=1 ; $i<=$arg_number; $i++)
		{
		my $name = $internal_branch->{'args'}->{$i}->{'name'};
		my $type = $internal_branch->{'args'}->{$i}->{'type'};
		my $settings = $internal_branch->{'args'}->{$i}->{'settings'};
		my $value = $internal_branch->{'args'}->{$i}->{'default'};
		my $hint = $internal_branch->{'args'}->{$i}->{'hint'};
		if ($style_group eq 'EMBOBJ')
			{
			my $object = $self->{'objects'}->{$tagname};
			$object->{'args'}->{$i}->{'name'}= $name;
			$object->{'args'}->{$i}->{'type'}= $type;
			$object->{'args'}->{$i}->{'settings'}= $settings;
			$object->{'args'}->{$i}->{'value'}=$value;
			$object->{'args'}->{$i}->{'hint'}=$hint;
			}
		elsif ($style_group eq 'DOC_STYLES')
			{
			$self->{'doc_args'}->{$i}->{'name'}= $name;
			$self->{'doc_args'}->{$i}->{'type'}= $type;
			$self->{'doc_args'}->{$i}->{'settings'}= $settings;
			$self->{'doc_args'}->{$i}->{'value'}=$value;
			$self->{'doc_args'}->{$i}->{'hint'}=$hint;
			}
		else
			{
			$self->{'args'}->{$tagname}->{$i}->{'name'}= $name;
			$self->{'args'}->{$tagname}->{$i}->{'type'}= $type;
			$self->{'args'}->{$tagname}->{$i}->{'settings'}= $settings;
			$self->{'args'}->{$tagname}->{$i}->{'value'}=$value;
			$self->{'args'}->{$tagname}->{$i}->{'hint'}=$hint;
			}
		}
	}
#Anzahl der Argumente zurueckliefern
return $arg_number;
}

=head3 _do_tag_click

$object->_do_tag_click($tagname)

Callback for handling the mouse click on a tag of a style that has arguments defined in 'edoc.int'.

Arguments:

(+) $tagname: The name of the tag which arguments should be shown

=cut

###############################
# Callback fuer Klick auf Tag #
###############################
sub _do_tag_click
{
my $self = shift;
my $tag_name = shift;
my $dlg=TagForm->new($self->docform(), $tag_name, 0, undef, 0);

}

1; #So muss package enden

=head1 KNOWN BUGS:

=head1 STILL MISSING:

$object ->get_par_styles_tags(); #Returns a list of all paragraph styles <PAR_STYLES> that are in the document

$object ->get_env_sytles_tags(); #Returns a list of all environment styles <ENV_STYLES> that are in the document

$object ->get_char_styles_tags(); #Returns a list of all character styles <CHAR_STYLES> that are in the document

$object ->get_command_format_tags(); #Returns a list of all commands <COMMANDS> that are in the document

$object ->get_embobj_tags(); #Returns a list of all embeded objects <EMBOBJ> that are in the document

Get/Set-methods for the not-simple attributes that are sub-hashes in the $self-hash:

(+) The $self->{'tags'} - branch

(+) The $self->{'args'} - branch

(+) The $self->{'env_numlists'} - branch

(+) The $self->{'doc_numlists'} - branch

(+) The $self->{'objects'} - branch

=cut

