#This source file belongs to e:doc
#Copyright (C) 1999  Thomas M. Schmickl
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License.
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#See the GNU General Public License for more details.
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Filesys;

use strict;

=head1 NAME:

Package: Filesys

=head1 SYNOPSIS:

use lib::Filesys;

=head1 DIRECTORY/FILE:

The file Filesys.pm is located in the subdirectory './lib'

=head1 DESCRIPTION:

This package contains all functions that read/write information stored in the block system that is used in the the configuration files
that are: 'edoc.cfg', 'edoc.int', 'edoc.obj', '{backend_name}.out', '{backend_name}.in', '{backend_name}.{object_name}.out' and '{backend_name}.{object_name}.in'.

Actually it does not perform directly on/to the file on disk, the files content has to be transferred into an array and a
reference to this filearray has then to be passed through the functions of this package lib::Filesys.

The functions in this packages can also modify this filearray, so that it can be saved again as a file later.

All information in the configuration files in kept in blocks, which can be nested.

All blocks start with a start-tag like <<<BEGIN::block_name>>> and end with an end-tag like <<<END::block_name>>>, where block_name is a placeholder for the actual blocks name.
Only the last started block can be closed with the end-tag, of course.

There are several types of blocks:

 #comments are started with an '#' at the beginning

 <<<BEGIN::WRAPPER>>> #this is the start-tag of a wrapper-block which only contains other blocks and has no own content

   <<<BEGIN::HASH_BLOCK>>> #this is the start-tag of a hash-block which contains a set of property/value pairs

     item1 = "something" #this is a text property

     item2 = 2 #this is a number

     item3 = "subitem1||subitem2||subitem3" #This is an list property where || is the items delimiter

   <<<END::HASH_BLOCK>>> #End-tag of the hash-block

   <<<BEGIN::ARRAY_BLOCK>>> #this is the start-tag of an array-block which contains lines that should be taken as an array

     somtext is here   

     and here

     and here too...

   <<<END::ARRAY_BLOCK>>> #End-tag of the array-block

 <<<END::WRAPPER>>> #End-tag of the wrapper block

You see, this is not far away from the basic XML structure and will be transformed sometimes into  XML.

=head1 STATUS:

Coding: 90%

Documentation: up-do-date


=head1 FUNCTIONS:

=head2 read_blocklist

@blocklist = Filesys::read_blocklist($ref_file_array)

Reads the whole filearray and produces a list (array) with all blocknames and returns it.

Parameters: 

(+) $ref_file_array: A reference to the array that holds the content of the configuration file

=cut

####################################################
# Liste aller vorhandenen Blcke in Datei erzeugen #
####################################################
sub read_blocklist
{
#Liest die im array orhandenen Blocknamen und liefert sie als Liste retour
my $ref_file_array = shift; #Referenz auf jenes Array, welches den Fileinhalt darstellt
my @block_list; #Liste aller Blcke welche sich auf der obersten Verschachtelungsebene befinden
my $in_block=0; #Count opened blocks
for (my $i=0; $i<=$#{$ref_file_array}; $i++)
	{
	my $line = $ref_file_array->[$i];
	$line =~ s/^\s*//o; #Remove whitespaces in front
	$line =~ s/[^\\]#.*\n$/\n/o ; #Remove comments that are not masked with a backslash
	$line =~ s/^#.*\n$/\n/o ; #Remove commands at the beginning of lines
	$line =~ s/\s*\n$/\n/o; #Remove whitespaces at the end of the line
	if ($line =~ /^<<</o) #Pre-checking this for speed issues (calling subroutines and returning values costs time)
		{
		my $check = Filesys::string_contains_tag($line);
		if (! $check == 0)
			{
			my $blockname = $check->{'block_name'};
			if ($check->{'tag_type'} eq 'start')
				{
				$in_block++ ;
				}
			if ($check->{'tag_type'} eq 'end')
				{
				if ($in_block == 1) #End of a toplevel block
					{
					push(@block_list,$blockname);
					}
				$in_block--;
				}
			}
		}
	}
return @block_list;
}

=head2 delete_block

$line = Filesys::delete_block($ref_file_array, $blockname)

Removes the whole block from the filearray 

Parameters: 

(+) $ref_file_array: A reference to the array that holds the content of the configuration file

(+) $blockname: The name of the block to remove

=cut

#############################################
# Einen Block lschen, sofern er existiert  #
#############################################
sub delete_block
{
#Lscht einen ganzen Block
my $ref_file_array = shift; #Referenz auf jenes Array, welches den Fileinhalt darstellt
my $blockname = shift; #Blockname der gelscht werden soll
my ($startfound,$endfound)=(0,0); #flags
my ($startpos,$endpos)=(-1,-1); #Positionen der Blockgrenzen
my $check=0; #flag
my @block_name_list = Filesys::read_blocklist;
foreach my $block (@block_name_list)
	{
	$check=1 if $block eq $blockname;
	}
if ($check==1)
	{
	my $range = Filesys::get_block_range($ref_file_array, $blockname);
	if (! $range==0)
		{
		splice(@$ref_file_array,$range->{'start_index'},$range->{'end_index'}-$range->{'start_index'}+1);
		}
	}
}

=head2 read_data_array

@block = Filesys::read_data_array($ref_file_array, $blockname, $process_edoc_commands)

Reads the content of a block as an array and returns it.

Parameters: 

(+) $ref_file_array: A reference to the array that holds the content of the configuration file

(+) $blockname: The name of the block to be read

(+) $process_edoc_commands: If set to 1, special edoc-commands like "%NL%" or "%INLINE_MACRO::....%" are processed.
    If set to 0, nothing is done. BEWARE: These processing will corrupt all hash-blocks that are inside this array-block !
    So this parameter is only set to 1 in the function that reads the macro-code, where no hash-block is inside.

=cut

###################################################
# Einen ganzen Block als Array von Datei einlesen #
###################################################
sub read_data_array
{
#Liest die Zeilen zwischen "<<<BEGIN::$block>>>" und "<<<END::$block>>>"
my $ref_file_array = shift; #Referenz auf jenes Array, welches den Fileinhalt darstellt
my $blockname = shift; #Name des Blockes in Datei
my $process_edoc_commands = shift;
$ process_edoc_commands=0 if not defined $process_edoc_commands;

my @return_array;
my $range = Filesys::get_block_range($ref_file_array, $blockname);
if (! $range==0)
	{
	if ($process_edoc_commands == 1)
		{
		for (my $i=$range->{'start_index'}+1;$i<$range->{'end_index'};$i++) {push(@return_array,Language::process_edoc_commands($ref_file_array->[$i]))}
		}
	else
		{
		for (my $i=$range->{'start_index'}+1;$i<$range->{'end_index'};$i++) {push(@return_array,$ref_file_array->[$i])}
		}
	}
return @return_array;
}

=head2 cut_data_array

@block = Filesys::cut_data_array($ref_file_array, $blockname, $process_edoc_commands)

Cuts the content of a block from an array and returns it.

Parameters: 

(+) $ref_file_array: A reference to the array that holds the content of the configuration file

(+) $blockname: The name of the block to be read

(+) $process_edoc_commands: If set to 1, special edoc-commands like "%NL%" or "%INLINE_MACRO::....%" are processed.
    If set to 0, nothing is done. BEWARE: These processing will corrupt all hash-blocks that are inside this array-block !
    So this parameter is only set to 1 in the function that reads the macro-code, where no hash-block is inside.

=cut

############################################
# Cut a whole array-block out of an array  #
############################################
sub cut_data_array
{
#Liest die Zeilen zwischen "<<<BEGIN::$block>>>" und "<<<END::$block>>>"
my $ref_file_array = shift; #Referenz auf jenes Array, welches den Fileinhalt darstellt
my $blockname = shift; #Name des Blockes in Datei
my $process_edoc_commands = shift;
$process_edoc_commands = 0 if not defined $process_edoc_commands;

my @return_array;
my $range = Filesys::get_block_range($ref_file_array, $blockname);

if (! $range==0)
	{
	if ($process_edoc_commands == 1)
		{
		for (my $i=$range->{'start_index'}+1;$i<$range->{'end_index'};$i++) {push(@return_array,Language::process_edoc_commands($ref_file_array->[$i]))}
		}
	else
		{
		for (my $i=$range->{'start_index'}+1;$i<$range->{'end_index'};$i++) {push(@return_array,$ref_file_array->[$i])}
		}
	my @dummy = splice(@{$ref_file_array},$range->{'start_index'},$range->{'end_index'}-$range->{'start_index'}+1);
	}

return @return_array;
}


=head2 write_data_array

Filesys::write_data_array($ref_file_array, $blockname, $ref_array_to_write, $no_delete_first)

writes the content of an array as an block into the filearray.

Parameters: 

(+) $ref_file_array: A reference to the array that holds the content of the configuration file

(+) $blockname: The name of the block to be read

(+) $ref_array_to_write: A reference to the array that should be written into the filearray

(+) $no_delete_first: If set to 1, the block is not seached first for deletion (if you know it does not exits)

=cut

#######################################################
# Einen ganzen Block als Array in die Datei schreiben #
#######################################################
sub write_data_array
{
#Schreibt das Array zwischen "<<<BEGIN::$block>>>" und "<<<END::$block>>>"
my $ref_file_array = shift; #Referenz auf jenes Array, welches den Fileinhalt darstellt
my $blockname = shift; #Name des Blockes in Datei
my $ref_array = shift; #Zeiger auf das zu schreibenede Array
my $no_delete_first = shift; #If set to 1 , no delete of the block first

if ((defined $no_delete_first)&&(! $no_delete_first==1))
	{
	Filesys::delete_block($ref_file_array,$blockname);
	}

my $starttag = Filesys::build_start_tag($blockname);
push (@$ref_file_array,$starttag);
foreach my $line (@$ref_array)
	{
	push (@$ref_file_array,$line);
	}
my $endtag = Filesys::build_end_tag($blockname);
push (@$ref_file_array,$endtag);
}

=head2 read_data_hash

%block = Filesys::read_data_hash($ref_file_array, $blockname)

Reads the content of a block as a hash and returns it.

Parameters: 

(+) $ref_file_array: A reference to the array that holds the content of the configuration file

(+) $blockname: The name of the block to be read

=cut

##################################################################
# Einen ganzen Block als Schlssel/Wert-Paare von Datei einlesen #
##################################################################
sub read_data_hash
{
#Liest die Zeilen zwischen "<<<BEGIN::$block>>>" und "<<<END::$block>>>"
my $ref_file_array = shift; #Referenz auf jenes Array, welches den Fileinhalt darstellt
my $blockname = shift; #Name des Blockes in Datei
my %return_hash;
my @an_array;
my $range = Filesys::get_block_range($ref_file_array, $blockname);
if (! $range==0)
	{
	for (my $i=$range->{'start_index'}+1;$i<$range->{'end_index'};$i++) {push(@an_array,$ref_file_array->[$i])}
	foreach my $line (@an_array)
		{
		if ($line =~ /^\s*(.+?)\s*=\s*(.*)\s*$/o)
			{
			my $key=$1;
			my $value=$2;
			#Unmask the string
			$value = Filesys::unmask_string($value);
			$key = Filesys::unmask_string($key);
			#Remove " at the beginning
			$value =~ s/^\s*"//o;
			#Remove " at the end
			$value =~ s/"\s*$//o;
			#build up the return-hash
			$return_hash{$key} = Language::process_edoc_commands($value);
			}
		}
	}
return %return_hash;
}

=head2 cut_data_hash

%block = Filesys::cut_data_hash($ref_file_array, $blockname)

Cuts the content of a block as a hash and returns it.

Parameters: 

(+) $ref_file_array: A reference to the array that holds the content of the configuration file

(+) $blockname: The name of the block to be read

=cut

##################################################################
# Cuts a whole hash-block out of the file_array and returns it   #
##################################################################
sub cut_data_hash
{
my $ref_file_array = shift; #Referenz auf jenes Array, welches den Fileinhalt darstellt
my $blockname = shift; #Name des Blockes in Datei
my %return_hash;
my @an_array;
my $range = Filesys::get_block_range($ref_file_array, $blockname);
if (! $range==0)
	{
	for (my $i=$range->{'start_index'}+1;$i<$range->{'end_index'};$i++) {push(@an_array,$ref_file_array->[$i])}
	foreach my $line (@an_array)
		{
		if ($line =~ /^\s*(.+?)\s*=\s*(.*)\s*$/o)
			{
			my $key=$1;
			my $value=$2;
			#Unmask the string
			$value = Filesys::unmask_string($value);
			$key = Filesys::unmask_string($key);
			#Remove " at the beginning
			$value =~ s/^\s*"//o;
			#Remove " at the end
			$value =~ s/"\s*$//o;
			#build up the return-hash
			$return_hash{$key} = Language::process_edoc_commands($value);
			}
		}
	my @dummy = splice(@{$ref_file_array},$range->{'start_index'},$range->{'end_index'}-$range->{'start_index'}+1);
	}
return %return_hash;
}


=head2 write_data_hash

Filesys::write_data_hash($ref_file_array, $blockname, $ref_hash_to_write, $no_delete_first)

writes the content of a ahsh as an block into the filearray.

Parameters: 

(+) $ref_file_array: A reference to the array that holds the content of the configuration file

(+) $blockname: The name of the block to be read

(+) $ref_hash_to_write: A reference to the hash that should be written into the filearray

(+) $no_delete_first: If set to 1, the block is not seached first for deletion (if you know it does not exits)

=cut

######################################################################
# Einen ganzen Block als Schlssel/Wert-Paare in die Datei schreiben #
######################################################################
sub write_data_hash
{
#Schreibt das Array zwischen "<<<BEGIN::$block>>>" und "<<<END::$block>>>"
my $ref_file_array = shift; #Referenz auf jenes Array, welches den Fileinhalt darstellt
my $blockname = shift; #Name des Blockes in Datei
my $ref_hash = shift; #Zeiger auf den zu schreibenede Hash
my $no_delete_first = shift; #If set to 1 , no delete of the block first

if ((defined $no_delete_first)&&(! $no_delete_first==1))
	{
	Filesys::delete_block($ref_file_array,$blockname);
	}

my $starttag = Filesys::build_start_tag($blockname);
push (@$ref_file_array,$starttag);

while ((my $key,my $value) = each %$ref_hash)
	{
	$value = '' if not defined $value;
	$value = Filesys::mask_string($value);
	$value =~ s/\n/%NL%/g; #Mask possible newlines into %NL%
	push (@$ref_file_array,"$key = \"".$value."\"");
	}

my $endtag = Filesys::build_end_tag($blockname);
push (@$ref_file_array,$endtag);

}

=head2 read_hash_value

$value = Filesys::read_hash_value($ref_file_array, $blockname, $key)

NOT CODED YET, but it should read one property entry in a hash-block.

Parameters: 

(+) $ref_file_array: A reference to the array that holds the content of the configuration file

(+) $blockname: The name of the block to be read

(+) $key: The name of the property entry (= key in the hash) 

=cut

###########################
# Einen Hasheintrag lesen #
###########################
sub read_hash_value
{
#Sucht nach dem Wert eines Schlssels in einem ganz bestimmten Block
my $ref_file_array = shift; #Referenz auf jenes Array, welches den Fileinhalt darstellt
my $blockname = shift; #Name des Blockes in Datei
my $key = shift; #Name des Schlssels

#CODE IS MISSING HERE, ITS NOT NEEDED YET

my $value;
return $value;
}

=head2 write_hash_value

Filesys::write_hash_value($ref_file_array, $blockname, $key, $value)

NOT CODED YET, but it should write one property / value pair in a hash-block.

Parameters: 

(+) $ref_file_array: A reference to the array that holds the content of the configuration file

(+) $blockname: The name of the block to be read

(+) $key: The name of the property entry (= key in the hash) 

(+) $value: The value of the property entry (= value in the hash) 

=cut

###########################
# Einen Hasheintrag lesen #
###########################
sub write_hash_value
{
#Sucht nach dem Wert eines Schlssels in einem ganz bestimmten Block
my $ref_file_array = shift; #Referenz auf jenes Array, welches den Fileinhalt darstellt
my $blockname = shift; #Name des Blockes in Datei
my $key = shift; #Name des Schlssels
my $value = shift; #Wert

#CODE IS MISSING HERE, ITS NOT NEEDED YET
}

=head2 split_stringlist

@a_list = Filesys::split_stringlist($source_string, $delimiter)

Splits a list-property into an array and returns it.

Parameters: 

(+) $source_string: The string that contains the list

(+) $delimiter: The delimiter used in the stringlist (Default: '||')

=cut

####################################################################################
# Eine Liste die mit einem Trenner als String codiert ist in eine Liste schreiben #
####################################################################################
sub split_stringlist
{
#Splittet einen String in eine Liste auf
my $string = shift; #bergebener String
my $trenner = shift; #Trennzeichen
$trenner = quotemeta($trenner);
my @list = split(/$trenner/,$string);
return @list;
}

=head2 is_in_stringlist

$bool = Filesys::is_in_stringlist($searched_string, $source_string, $delimiter)

Checks if the searched string is in the stringlist and returns 1 if it is there and 0 if its not there.

Parameters: 

(+) $searched_string: The string that is searched in the stringlist

(+) $source_string: The string that contains the list

(+) $delimiter: The delimiter used in the stringlist (Default: '||')

=cut

################################################
# Prft ob ein String in einer Stringliste ist #
################################################
sub is_in_stringlist
{
#Splittet einen String in eine Liste auf
my $searchstring = shift; #bergebener Suchstring
my $stringlist = shift; #Stringliste , die durchsucht werden soll
my $trenner = shift; #Trennzeichen
my $result=0;
my @list = split_stringlist($stringlist,$trenner);
for (my $i=0; $i<=$#list; $i++)
	{
	if ($searchstring eq $list[$i])
		{
		$result=1;
		}
	}
return $result;
}

=head2 check

$bool = Filesys::check($ref_file_array)

Checks if the fileformat in the filearray is correct, that means if all blocks are opened and closed correctly.

Returns -1 (<---Attention!!!) if the filearray is correct otherwise the error message is printed to STOUT and the number of the line
that contains the error is returned by the function.

Parameters: 

(+) $ref_file_array: A reference to the array that holds the content of the configuration file

=cut

#####################################################
# berprft auf korrkete Verschachtelung der Blcke #
#####################################################
sub check
{
my $ref_file_array = shift; #Referenz auf jenes Array, welches den Fileinhalt darstellt
my @stack; #Stack fr Blocknamen
my $error=-1; #Flag fr Korrektheit (bei -1 alles ok)
for (my $i=0; $i<=$#{$ref_file_array}; $i++)
	{
	my $line = $ref_file_array->[$i];
	if ($line =~ /^<<</o) #Pre-checking this for speed issues (calling subroutines and returning values costs time)
		{
		my $check = Filesys::string_contains_tag($line);
		if (! $check == 0)
			{
			my $blockname = $check->{'block_name'};
			if ($check->{'tag_type'} eq 'start')
				{
				push(@stack,quotemeta($blockname)); #Blockname  auf Stack legen
				}
			if ($check->{'tag_type'} eq 'end')
				{
				if ($#stack>=0)
					{
					my $act_block = pop(@stack); #Letzten begonnenen Blocknamen rausholen
					if (($act_block eq quotemeta($blockname))==0) #Bei Ungleichheit
						{
						$error=$i; #Fehler in Verschachtelung, Zeilennummer retourliefern
						print "I expected end of block $act_block, but found end of block $blockname instead.\n";  
						last; #Schleife abbrechen
						}
					}
				else #Wenn END gefunden wurde obwohl Stack bereits leer ist
					{
					print "Begin of Block $blockname not found in file.\n";
					$error=$i; #Fehler in Verschachtelung, Zeilennummer retourliefern
					last; #Schleife abbrechen
					}
				}
			}
		}
	}
if (($#stack>=0)&&($error==-1))#Wenn stack noch etwas drinnenhat und bis jetzt noch kein Fehler Entleerung verhindert hat 
	{
	$error=$#{$ref_file_array}; #Zeilennummer auf letzte Zeile
	print "At least last block was not closed.\n";
	} 
return $error;
}

=head2 expand_include

$found_include = Filesys::expand_include($ref_file_array)

Expand the referenced filearray by inlining the files that are mentioned inside of an "<<<INCLUDE_FILE::....>>>" statement.

This function returns 0 if no INCLUDE statement was found or the number of the (recursively) found INCLUDE statement

Parameters: 

(+) $ref_file_array: A reference to the array that holds the content to be expanded

=cut

#############################################
# Expand an <<<INCLUDE_FILE::....>>>-block  #
#############################################
sub expand_include
{
#Lscht einen ganzen Block
my $ref_file_array = shift; #Referenz auf jenes Array, welches den Fileinhalt darstellt
my $found_include=0;
for (my $i=0;$i<=$#{$ref_file_array};$i++)
	{
	my $line = $ref_file_array->[$i];
	if ($line =~ /^<<<INCLUDE_FILE::(.*)>>>/o)
		{
		my @include_array;

		#Check for the file and read it into an include_array (The existance is checked there)
		my $filename = Utils::process_filename($1);
		if (defined $filename)
			{
			@include_array = InOut::build_array_from_file($filename);
			}
		else
			{
			die "Could not find file \"$filename\" for including it (Used by an <<<INCLUDE_FILE::...>>>-block !\n";
			}
		#and delete the $line from the original array
		#and include the include_array there
		splice(@{$ref_file_array},$i,1,@include_array);
		}
	}
return $found_include;
}

=head2 mask_string

$resut_string = Filesys::mask_string($source_string)

Masks the given string. Characters that are special for e:doc like ", <, >, :, %, #, |, \, = are masked with a backslash before it.

This function returns the result string.

Parameters: 

(+) $source_string: The string to be masked

=cut

#########################################
# Unmask a string for e:doc filesystem  #
#########################################
sub mask_string
{
my $string = shift; #The given string to mask
$string =~ s/(["<>:%=#\|\\])/\\$1/go ;
return $string;
}

=head2 unmask_string

$resut_string = Filesys::unmask_string($source_string)

Unmasks the given string. masked characters like \" \< \> \: \% \# \| \= \\ are unmasked again.

This function returns the result string.

Parameters: 

(+) $source_string: The string to be masked

=cut

#########################################
# Unmask a string for e:doc filesystem  #
#########################################
sub unmask_string
{
my $string = shift; #The given string to mask
$string =~ s/\\(["<>:%=#\|\\])/$1/go ;
return $string;
}

=head2 build_start_tag

$tag = Filesys::build_start_tag($block_name);

Builds up the start-tag of the given block.

This function returns the tag.

Parameters: 

(+) $block_name: The name of the block

=cut

##################################
# Create a start-tag for a block #
##################################
sub build_start_tag
{
my $block_name = shift; #The given blockname
return '<<<BEGIN::'.$block_name.'>>>';
}

=head2 build_end_tag

$tag = Filesys::build_end_tag($block_name);

Builds up the end-tag of the given block.

This function returns the tag.

Parameters: 

(+) $block_name: The name of the block

=cut

################################
# Create a end-tag for a block #
################################
sub build_end_tag
{
my $block_name = shift; #The given blockname
return '<<<END::'.$block_name.'>>>';
}

=head2 string_contains_tag

$tag = Filesys::string_contains_tag($string);

Checks for the existance of a tag in the given string.

Returns 0 if no tag is inside the string or a reference to an anonymous hash that holds the following
result parameters:

$result->{'block_name'} contains the name of the block that was found

$result->{'tag_type'} = 'start' if its an start-tag

$result->{'tag_type'} = 'end' if its an end-tag

Parameters:

(+) $string: The string to be searched for a tag

=cut

#################################
# Check for a tag in the string #
#################################
sub string_contains_tag
{
my $a_string = shift; #The string to be seached
my $result;
if ($a_string =~ /^<<<(.*)::(.*)>>>$/o)
	{
	if (($1 eq 'BEGIN')||($1 eq 'END')) #if its a correct start- or end-tag
		{
		$result = {}; #Make $result to an reference to an hash
		$result->{'block_name'} = $2; #store the blockname
		if ($1 eq 'BEGIN') #if its a start tag
			{
			$result->{'tag_type'} = 'start';
			}
		else #else it has to be an end-tag
			{
			$result->{'tag_type'} = 'end';
			}
		}
	else
		{
		$result = 0; #No start- or end-tag found
		}
	}
else
	{
	$result=0;#No tag found
	}
return $result;
}

=head2 get_block_range

$tag = Filesys::get_block_range($ref_filearray, $blockname);

Gets the arrayindexes of the lines the block spans.

Returns 0 if no tag is inside the string or a reference to an anonymous hash that holds the following
result parameters:

$result->{'start_index'} contains startindex of the block

$result->{'end_index'} contains endindex of the block

Parameters:

(+) $ref_filearray: A reference to the filearray

(+) $blockname: The name of the block

=cut

###########################################
# Get the start and endindex of the block #
###########################################
sub get_block_range
{
my $ref_file_array = shift;
my $searched_blockname = shift;
my $result; # A scalar return value
my ($startfound,$endfound)=(0,0); #flags
my ($startpos,$endpos)=(-1,-1); #Positionen der Blockgrenzen
my $level=0; #stack

for (my $i=0;$i<=$#{$ref_file_array};$i++)
	{
	my $line = $ref_file_array->[$i];
	if ($line =~ /^<<</o) #Pre-checking this for speed issues (calling subroutines and returning values costs time)
		{
		my $check = Filesys::string_contains_tag($line);
		if (! $check == 0) #If there is a tag in the line
			{
			my $found_blockname = $check->{'block_name'};
			if ($check->{'tag_type'} eq 'start')
				{
				if (($level==0)&&($found_blockname eq $searched_blockname))
					{
					$startfound=1;
					$startpos=$i;
					}
				$level++
				}
			if ($check->{'tag_type'} eq 'end')
				{
				if (($level==1)&&($found_blockname eq $searched_blockname))
					{
					$endfound=1;
					$endpos=$i;
					last;
					}
				$level--;
				}
			}
		}
	}
if (($startfound==1)&&($endfound==1)&&($endpos>$startpos)) #if both tags have been found
	{
	$result = {}; #Make $result tp an reference of an hash
	$result->{'start_index'} = $startpos; #return startindex
	$result->{'end_index'} = $endpos; #return endindex
	}
else
	{
	$result = 0; #return 0
	}
return $result;
}


1; #So mu package enden

=head1 KNOWN BUGS:

=head1 STILL MISSING:

=cut
