#This package is created by: Thomas Schmickl, Graz, Austria.
#Copyright by: Thomas Schmickl, Graz, Austria.
#License: This code has the same license conditions like Perl itself.
package XGrid;

require Tk::Frame;
require Tk::Text;
require Tk::Photo;
require Tk::Label;
require Tk::Pane;
require Tk::Balloon;
use Tk qw(Ev);

$VERSION = '0.01';

@ISA = qw(Tk::Derived Tk::Frame);
Tk::Widget->Construct('XGrid');

use strict;

#DEFAULT SETTINGS
my $default_cell_width = 120; #The standard width of a cell in pixels (if you do not specify it)
my $default_cell_height = 40; #The standard height of a cell in pixels (if you do not specify it)
my @styledefs_text_tag = ('-justify', '-font'); #Options for the tag displayed in the Text widget
my @styledefs_text_widget = ('-foreground', '-relief', '-borderwidth', '-font'); #Options for the Text widget itself
my @styledefs_label_widget = ('-background', '-foreground', '-relief', '-borderwidth', '-font', '-justify', '-anchor'); #Options for the Label widget itself

my $table_xpm = <<END_TABLE_XPM;
/* XPM */
static char*v_handle[]={
"18 18 5 1",
"a c #d7d6d6",
"b c #9b9a9a",
". c None",
"c c #020202",
"# c #ffffff",
"..................",
"..................",
"..................",
"........##........",
".......#aa#.......",
".......#aa#.......",
".......#aa#.......",
"....####aa####....",
"...#aaaaccaaaa#...",
"...#aaaaccaaaa#...",
"....####aa####....",
".......#aa#.......",
".......#aa#.......",
".......#aa#.......",
"........##........",
"..................",
"..................",
".................."};
END_TABLE_XPM

my $h_handle_xpm = <<END_H_HANDLE_XPM;
/* XPM */
static char*h_handle[]={
"12 18 5 1",
"b c #d7d6d6",
". c #c0c0c0",
"# c #9b9a9a",
"c c #020202",
"a c #ffffff",
"..#...#..#..",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"abc.abc.abc.",
"..#...#..#.."};
END_H_HANDLE_XPM

my $v_handle_xpm = <<END_V_HANDLE_XPM;
/* XPM */
static char*v_handle[]={
"18 12 5 1",
"a c #d7d6d6",
"b c #9b9a9a",
". c None",
"c c #020202",
"# c #ffffff",
".################.",
".aaaaaaaaaaaaaaaa.",
"bccccccccccccccccb",
"..................",
".################.",
".aaaaaaaaaaaaaaaa.",
"bccccccccccccccccb",
"..................",
".################.",
"baaaaaaaaaaaaaaaab",
".cccccccccccccccc.",
".................."};
END_V_HANDLE_XPM

=head1 NAME:

Widget: XGrid

=head1 SYNOPSIS:

use Tk::XGrid( -initialrows=>value, -initialcols=>value ); 

=head1 DIRECTORY/FILE/INSTALLATION:

Just drop the file and the 3 xpm-images into the Tk directory (_perldir_/lib/site/Tk).

=head1 DESCRIPTION:

This widget is a little spreadsheet-like grid of texts.

It was developed for working as a embeded table in my text-processor project 'e:doc'. It
offers a set of Labels that are organised in rows and column yousing the grid layout manager. 

If you click on a cell, the Label is replaced by a Text widget for editing the cells content.

You can resize the columns and the rows by dragging the resize-grips with the mouse.

You can define multiple 'cell-styles' that can be applied to each cell. This way you can
change properties like 'background', 'foreground', 'relief', 'font' etc.

=head1 DEFINITIONS:

=head2 Indexes:

The column- and row-indexes are integers between 1 and $widget->cols() or $widget->rows().

=head2 Cell styles:

If you do not define own styles, there will exist a style 'BASE' that is used for all cells.

If you want to change some settings of this style 'BASE', overwrite them using the method "define_style'.

If you do define other styles too AND want a changed style 'BASE', redefine the style 'BASE' first.

=head1 STATUS:

Coding: 80% done

Documentation: up-to-date

=head1 ATTRIBUTES:

Do not set/read them directly, use the Get/Set methods instead.

[... Attribute list missing here ...]

=head1 METHODS:

=head2 Class Methods:

=head3 ClassInit($class, $parentwindow)


The initialisation of  the class XGrid and calls ClassInit() of the SUPER class (Frame). 
Maybe I should create the images here and not in Populate().

=cut

##################
# Init the class #
##################
sub ClassInit
{
my ($class,$mainwindow)=@_; # Klassenname und Mainwindow
$class->SUPER::ClassInit($mainwindow); # Aufruf des Vorfahren
return $class;
}

=head2 Constructors:

=head3 Populate

$object = XGrid(%options); 

This 'constructor' sets the options of the new object of the class XGrid. 

The binding of events and the creation of the associated Baloonhelp is done here too.

The keys/values in %options: 

(+) '-initialrows': The number of rows that should be there initially [Only evaluated at widget creation time, later changes are useless].

(+) '-initialcols': The number of cols that should be there initially [Only evaluated at widget creation time, later changes are useless].

=cut

######################################
# Evaluating the -option/value pairs #
######################################
sub Populate
{
my ($self,$args)=@_; 

my $initial_cols;
my $initial_rows;

$initial_cols = delete ($args->{'-initialcols'} );
$initial_rows = delete ($args->{'-initialrows'} );

$self->SUPER::Populate($args); #Calling the SUPER class

#The default values for the number of actual cols and rows (they are 0, because nothing is created yet)
$self->cols(0);
$self->rows(0);


#Building up the table itself
$self->{'_table'} = $self->Scrolled('Pane', -scrollbars=>'osoe', -sticky=>'nw');

$self->{'_col_grip_img'} = $self->Pixmap(-data=>$h_handle_xpm);
$self->{'_row_grip_img'} = $self->Pixmap(-data=>$v_handle_xpm);
$self->{'_table_img'} = $self->Pixmap(-data=>$table_xpm);
$self->update();

#At the beginning, no cell is active
$self->{'_active_cell_col'} = 0;
$self->{'_active_cell_row'} = 0;

#Preparing the Text widget for entries
$self->{'_text_entry'} = $self->{'_table'}->Scrolled('Text', -scrollbars=>'osoe', -width=>1, -height=>1, -wrap=>'word', -background=>'yellow');
$self->{'_text_entry'}->GeometryRequest(10,10);
$self->{'_text_entry'}->bind('<Button-3>' => sub{$self->_cell_popup( Ev('X'), Ev('Y'), $self, $self->{'_active_cell_col'}, $self->{'_active_cell_row'})});
$self->{'_text_entry'}->bind('<Key>'=> sub{shift->tagAdd($self->{'_cell_styles'}[$self->{'_active_cell_col'}][$self->{'_active_cell_row'}], '1.0', 'end')});
$self->{'_text_entry'}->bind('<Control-Key-Down>', [\&_nav_key_released, $self, 'down']);
$self->{'_text_entry'}->bind('<Control-Key-Up>', [\&_nav_key_released, $self, 'up']);
$self->{'_text_entry'}->bind('<Control-Key-Right>', [\&_nav_key_released, $self, 'right']);
$self->{'_text_entry'}->bind('<Control-Key-Left>', [\&_nav_key_released, $self, 'left']);
$self->{'_text_entry'}->bind('<Control-Key-Home>', sub{$self->_show_left_upper_cell() } );
$self->{'_text_entry'}->bind('<Control-Key-End>', sub{$self->_show_right_bottom_cell() } );
$self->{'_text_entry'}->menu(undef);

#Defining the style 'BASE' (DO NOT REMOVE, NOT EVEN ONE OF THE OPTIONS)
my $style_options = {'-background'=>'white', '-foreground'=>'black', '-relief'=>'ridge', '-borderwidth'=>2, '-justify'=>'center', '-anchor'=>'n', '-font'=>'-adobe-times-medium-r-normal-*-*-160-*-*-p-*-*'};
$self->define_style('BASE', $style_options);

#Building up the upper-left table button
$self->{"_table_button"} = $self->{'_table'}->Frame(-width=>54, -height=>22, -borderwidth=>2, -relief=>'raised')->grid(-row=>0, -column=>0,  -sticky=>'nsew');
$self->{"_table_button"}->packPropagate(0);
$self->{'_table_button_label'} = $self->{"_table_button"}->Label(-image=>$self->{'_table_img'})->pack(-expand=>1, fill=>'both');
$self->_create_balloonhelp($self->{'_table_button_label'}, 'Right mouse button shows table-commands', 'balloon', undef);
$self->{'_table_button_label'}->bind('<Button-3>' => [\&_table_popup, Ev('X'), Ev('Y'), $self] );

#Building up the column headers
while ($self->cols() < $initial_cols)
	{
	$self->add_col($default_cell_width);
	}

#Building up the rows
while ($self->rows() < $initial_rows)
	{
	$self->add_row($default_cell_height);
	}

$self->{'_table'}->pack(-expand=>1, -fill=>'both');

$self->update();

return $self;
}


=head2 Destructors:

=head3 DESTROY 

The standard automatically called PERL destructor for objects.

=cut

#############################
# Automatic Perl destructor #
#############################
sub DESTROY
{
my $self = shift;

}

=head2 Public data methods (=Get/Set methods):

=cut

##############################
### PUBLIC GET/SET METHODS ###
##############################

=head3 _rows

$rows = $object->rows(); #gets the number of rows 

$object->rows($rows); #sets the number of rows (Only for internal use, to add rows you can only use the methods 'add_row' and 'insert_row')

=cut

### rows ###
sub rows
{
my $self=shift;
@_ ? $self->{'rows'} = shift : return $self->{'rows'};
}

=head3 cols

$cols = $object->cols(); #gets the number of columns

$object->cols($cols); #sets the number of columns (Only for internal use, to add columns you can only use the methods 'add_col' and 'insert_col')

=cut

### cols ###
sub cols
{
my $self=shift;
@_ ? $self->{'cols'} = shift : return $self->{'cols'};
}

=head2 Other public methods:

=head3 add_row

@array_of_widgets = $object->add_row( [$row_height] ); 

This function adds a row at the bottom of the table.

The function returns an array of widget-references. The first element is a
reference to the row-header (class "Frame"), the other elements are the
cell-widgets (class "Text") starting from the very left to the right.

Parameters:

(+) $row_height: The height of the new row

=cut

###############################################
# Adding a new row at the bottom of the table #
###############################################
sub add_row
{
my $self = shift;
my $row_height = shift;

$row_height = $default_cell_height if not defined $row_height;

my $table = $self->{'_table'};

my @return_array = ();

#Increase the number of rows
my $row = $self->rows();
$row++;
$self->rows($row);

#Produce the row header without gridding it
my $header = $self->{'_row_headers'}[$row] = $table->Frame(-relief=>'raised', -borderwidth=>2, -height=>$row_height);
$header->packPropagate(0);
my $grip = $header->Label(-image=>$self->{'_row_grip_img'}, -anchor=>'s', -justify=>'center', -cursor=>'hand2')->pack(-side=>'bottom', -anchor=>'s', -fill=>'x');
$self->_create_balloonhelp($grip, 'Resizing the row', 'balloon', undef);
my $label = $header->Label(-text=>"Row $row", -anchor=>'center', -justify=>'center')->pack(-expand=>1, fill=>'both');
$self->_create_balloonhelp($label, 'Right mouse button shows row-commands', 'balloon', undef);
$label->bind('<Button-3>' => [\&_row_heading_popup, Ev('X'), Ev('Y'), $self, $row]);
$grip->bind('<ButtonPress-1>', [\&_row_heading_mouse_button1_pressed, $self, Ev('X'), Ev('Y'), $row]);
$grip->bind('<ButtonRelease-1>', [\&_row_heading_mouse_button1_released, $self, Ev('X'), Ev('Y'), $row]);
$self->{'_row_heights'}[$row] = $row_height;

#Configure the new row
$table->gridRowconfigure($row, -minsize=>1);

#Creating the cells of the whole row without gridding them
for (my $x = 1; $x <= $self->cols(); $x++)
	{
	my $col_width = $self->{'_col_widths'}[$x];
	my $entry = $self->_create_cell($x, $row, $col_width , $row_height, 0); 
	#Store the created Entry in the widget array
	push (@return_array, $entry);
	}

#Gridding the whole row at once
$header->grid(@return_array, -sticky=>'nsew');

#Storing the header at the beginning of the return array
unshift (@return_array, $header);

$self->{'_table'}->update();

#Return the widget-array
return @return_array;

}

=head3 add_col

@array_of_widgets = $object->add_col( [$col_width] ); 

This function adds a column at the very right of the table.

The function returns an array of widget-references. The first element is a
reference to the column-header (class "Frame"), the other elements are the
cell-widgets (class "Text") starting from the top to the bottom.

Parameters:

(+) $col_width: The width of the new column

=cut

#################################################
# Adding a new column at the right of the table #
#################################################
sub add_col
{
my $self = shift;
my $col_width = shift;

$col_width = $default_cell_width if not defined $col_width;

my $table = $self->{'_table'};

my @return_array = ();

#Increase the number of rows
my $col = $self->cols();
$col++;
$self->cols($col);

#Produce the column header and grid it
my $header = $self->{'_col_headers'}[$col] = $table->Frame(-relief=>'raised', -borderwidth=>2, -width=>$col_width)->grid(-row=>0, -column=>$col, -sticky=>'nsew');
$header->packPropagate(0);
my $grip = $header->Label(-image=>$self->{'_col_grip_img'}, -cursor=>'hand2')->pack(-side=>'right', -anchor=>'e', -expand=>0, -fill=>'y');
$self->_create_balloonhelp($grip, 'Resizing the column', 'balloon', undef);
my $label = $header->Label(-text=>"Col $col", -justify=>'center', -anchor=>'center')->pack(-side=>'left', -anchor=>'w', -expand=>1, -fill=>'both');
$self->_create_balloonhelp($label, 'Right mouse button shows column-commands', 'balloon', undef);
$label->bind('<Button-3>' => [\&_col_heading_popup, Ev('X'), Ev('Y'), $self, $col]);
$grip->bind('<ButtonPress-1>', [\&_col_heading_mouse_button1_pressed, $self, Ev('X'), Ev('Y'), $col]);
$grip->bind('<ButtonRelease-1>', [\&_col_heading_mouse_button1_released, $self, Ev('X'), Ev('Y'), $col]);
$self->{'_col_widths'}[$col] = $col_width;

push (@return_array, $header);

#Configure the new column
$table->gridColumnconfigure($col, -minsize=>1);

#Creating the cells of the whole column and grid them
for (my $y = 1; $y <= $self->rows(); $y++)
	{
	my $row_height = $self->{'_row_heights'}[$y];
	my $entry = $self->_create_cell($col, $y, $col_width, $row_height, 1); 
	#Store the created Entry in the widget array
	push (@return_array, $entry);
	}

$self->{'_table'}->update();

#Return the widget-array
return @return_array;
}

=head3 activate_cell

$object->activate_cell($col, $row);

This method activates the specified cell and deactivates the former
active cell (if there was an active cell before).

Parameters:

(+) $col: The column that contains the cell

(+) $row: The row that contains the cell

=cut

###################
# Activate a cell #
###################
sub activate_cell
{
my $self = shift;
my $col = shift;
my $row = shift;

if ( defined $self->_test_cell($col, $row, 'activate_cell' ) )
	{
	_activate_cell(undef, $self, $col, $row);
	}
}

=head3 deactivate_cell

$object->deactivate_cell([$col, $row]);

This method deactivates the specified cell if its currently activated.

If both parameters are undef or omitted, the current active cell will be deactivated.

Parameters:

(+) $col: The column that contains the cell

(+) $row: The row that contains the cell

=cut

#####################
# Deactivate a cell #
#####################
sub deactivate_cell
{
my $self = shift;
my $col = shift;
my $row = shift;

if ($self->{'_active_cell_col'} == 0 and $self->{'_active_cell_row'} == 0)
	{
	goto EXIT;
	}
if ( not defined $col and not defined $row )
	{
	_deactivate_cell(undef, $self, $self->{'_active_cell_col'}, $self->{'_active_cell_row'} );
	}
elsif ( defined $self->_test_cell($col, $row, 'deactivate_cell' ) )
	{
	_deactivate_cell(undef, $self, $col, $row);
	}
EXIT:
}

=head3 put

$object->put($col, $row, $text);

Put some text into a cell

Parameters:

(+) $col: The column that contains the cell

(+) $row: The row that contains the cell

(+) $text: The text to be displayed there

=cut

#############################
# Put some text into a cell #
#############################
sub put
{
my $self = shift;
my $col = shift;
my $row = shift;
my $text = shift;

if ( defined $self->_test_cell($col, $row, 'put' ) )
	{
	if ($self->{'_active_cell_col'} == $col and $self->{'_active_cell_row'} == $row)
		{
		$self->{'_text_entry'}->delete('1.0', 'end');
		$self->{'_text_entry'}->insert('1.0', $text);
		}
	else
		{
		$self->{'_cell_contents'}[$col][$row] = $text;
		}
	}
}

=head3 get

my $text = $object->put($col, $row);

Get the content (text) of a cell.

The function returns the text of the cell.

Parameters:

(+) $col: The column that contains the cell

(+) $row: The row that contains the cell

=cut

#############################
# Get some text from a cell #
#############################
sub get
{
my $self = shift;
my $col = shift;
my $row = shift;

my $text = undef;

if ( defined $self->_test_cell($col, $row, 'get' ) )
	{
	if ($self->{'_active_cell_col'} == $col and $self->{'_active_cell_row'} == $row)
		{
		$text = $self->{'_text_entry'}->get('1.0', 'end');
		chomp($text);
		}
	else
		{
		$text = $self->{'_cell_contents'}[$col][$row];
		}
	}
return $text;
}

=head3 define_style

$object->define_style($style_name, $options);

Defines a new style for a cell.


Parameters:

(+) $style_name: The name for the style. If there is already a style with this name,
it's overwritten.

(+) $options: A reference to a hash holding the options for that style.

Possible options are: '-background', '-foreground', '-relief', '-borderwidth', '-font', '-justify'[, '-anchor'(this is ignored in the active-mode)]

If you do not set an option explecitly, the default option from the default style 'BASE' will be taken.

If you want other default options, just overwrite 'BASE' as the first style.

=cut

######################
# Define a new style #
######################
sub define_style
{
my $self = shift;
my $style_name = shift;
my $options = shift;

#Copy values form style 'BASE'
if (not ($style_name eq 'BASE'))
	{
	#Copy the settings for the text tags
	foreach my $key ( keys %{$self->{'_style_defs'}->{'BASE'}->{'text_tag'}} )
		{
		$self->{'_style_defs'}->{$style_name}->{'text_tag'}->{$key} = $self->{'_style_defs'}->{'BASE'}->{'text_tag'}->{$key};
		}
	#Copy the settings for the text widget
	foreach my $key ( keys %{$self->{'_style_defs'}->{'BASE'}->{'text_widget'}} )
		{
		$self->{'_style_defs'}->{$style_name}->{'text_widget'}->{$key} = $self->{'_style_defs'}->{'BASE'}->{'text_widget'}->{$key};
		}
	#Copy the settings for the label widget
	foreach my $key ( keys %{$self->{'_style_defs'}->{'BASE'}->{'label_widget'}} )
		{
		$self->{'_style_defs'}->{$style_name}->{'label_widget'}->{$key} = $self->{'_style_defs'}->{'BASE'}->{'label_widget'}->{$key};
		}
	}

#Overwrite it with the new values
foreach my $key ( keys %{$options} )
	{
	#Overwriting with the new options for text tags
	foreach my $item (@styledefs_text_tag)
		{
		if ($item eq $key)
			{
			$self->{'_style_defs'}->{$style_name}->{'text_tag'}->{$key} = $options->{$key};
			}
		}
	#Overwriting with the new options for the text widget
	foreach my $item (@styledefs_text_widget)
		{
		if ($item eq $key)
			{
			$self->{'_style_defs'}->{$style_name}->{'text_widget'}->{$key} = $options->{$key};
			}
		}
	#Overwriting with the new options for the label widgets
	foreach my $item (@styledefs_label_widget)
		{
		if ($item eq $key)
			{
			$self->{'_style_defs'}->{$style_name}->{'label_widget'}->{$key} = $options->{$key};
			}
		}
	}

#Teaching the Textwidget the tags
my %text_tag_options = %{$self->{'_style_defs'}->{$style_name}->{'text_tag'}} ;
$self->{'_text_entry'}->tagConfigure($style_name, %text_tag_options);
}

=head3 set_cell_style

$object->set_cell_style($col, $row, $style_name);

This method changes the cell to the given style.

Parameters:

(+) $col: The column that contains the cell

(+) $row: The row that contains the cell

(+) $style_name: The name of the style

=cut

#####################
# Change cell style #
#####################
sub set_cell_style
{
my $self = shift;
my $col = shift;
my $row = shift;
my $style_name = shift;

#Testing if the cell exists
if (defined $self->_test_cell($col, $row, 'set_cell_style') )
	{
	#testing if the style exists
	if (exists $self->{'_style_defs'}->{$style_name})
		{
		#testing if the Text widget is currently displayed in this cell
		if ($col == $self->{'_active_cell_col'} and $row == $self->{'_active_cell_row'})
			{
			#configuring the current tag and the text widget itself
			$self->_configure_text_entry($style_name);
			}
		#Storing the stylename as the current stylename for this cell
		$self->{'_cell_styles'}[$col][$row] = $style_name;
		#Configure the Label that belongs to the cell
		$self->_configure_label_entry($col, $row, $style_name);
		}
	else
		{
		die "ERROR (XGrid.pm): Style '$style_name' is not defined !\n";
		}
	}
}

=head3 get_cell_style

my $style_name = $object->get_cell_style($col, $row);

This method retruns the style of the cell.

Parameters:

(+) $col: The column that contains the cell

(+) $row: The row that contains the cell

=cut

##################
# get cell style #
##################
sub get_cell_style
{
my $self = shift;
my $col = shift;
my $row = shift;

#The default style
my $style_name = 'BASE';

#Testing if the cell exists
if (defined $self->_test_cell($col, $row, 'get_cell_style') )
	{
	#testing if the style exists
	$style_name = $self->{'_cell_styles'}[$col][$row];
	}

return $style_name;
}


=head3 style_list

my @style_names = $object->style_list();

Returns an array of all defined styles in the table

=cut

##################
# Get style list #
##################
sub style_list
{
my $self = shift;

my @style_list = keys %{$self->{'_style_defs'}};

return @style_list;
}

=head3 delete_style

$object->delete_style($style_name);

Deletes the style from the list of available styles. All settings for that
style are deleted as well. Cells containing that style are changed to style 'BASE'.

=cut

##################
# Delete a style #
##################
sub delete_style
{
my $self = shift;
my $style_name = shift;

if (not ($style_name eq 'BASE') and exists $self->{'_style_defs'}->{$style_name})
	{
	delete ($self->{'_style_defs'}->{$style_name});
	for (my $col = 1; $col <= $self->cols(); $col++)
		{
		for (my $row = 1; $row <= $self->rows(); $row++)
			{
			if ($style_name eq $self->{'_cell_styles'}[$col][$row])
				{
				$self->set_cell_style($col, $row, 'BASE');
				}
			}
		}
	}
}

=head3 get_col_width

my $width = $object->get_column_width($col);

Returns the actual width of the column.

Parameters:

(+) $col: The column index

=cut

####################
# Get column width #
####################
sub get_col_width
{
my $self = shift;
my $col = shift;

my $width = undef;

if ($col > $self->cols() or $col <1)
	{
	die "ERROR (XGrid.pm): Invalid column index. \n"; #debug
	}
else
	{
	$width = $self->{'_col_widths'}[$col];
	}
return $width;
}

=head3 set_col_width

$object->set_column_width($col, $width);

Changes the column to this width (in pixels).

Parameters:

(+) $col: The column index

(+) $width: The width the column should have

=cut

####################
# Set column width #
####################
sub set_col_width
{
my $self = shift;
my $col = shift;
my $width = shift;

if ($col > $self->cols() or $col <1)
	{
	die "ERROR (XGrid.pm): Invalid column index. \n"; #debug
	}
else
	{
	$self->{'_table'}->idletasks;
	my $heading_height = $self->{'_col_headers'}[$col]->Height();
	$self->{'_col_headers'}[$col]->GeometryRequest($width,$heading_height);
	$self->{'_col_widths'}[$col] = $width;
	$self->{'_table'}->idletasks;
	}
}

=head3 get_row_height

my $height = $object->get_row_height($row);

Returns the actual height of the row.

Parameters:

(+) $row: The row index

=cut

##################
# Get row height #
##################
sub get_row_height
{
my $self = shift;
my $row = shift;

my $height = undef;

if ($row > $self->rows() or $row <1)
	{
	die "ERROR (XGrid.pm): Invalid row index. \n"; #debug
	}
else
	{
	$height = $self->{'_row_heights'}[$row];
	}
return $height;
}

=head3 set_row_height

$object->set_row_height($row, $height);

Changes the row to this height (in pixels).

Parameters:

(+) $row: The row index

(+) $height: The height the row should have

=cut

##################
# Set row height #
##################
sub set_row_height
{
my $self = shift;
my $row = shift;
my $height = shift;

if ($row > $self->rows() or $row <1)
	{
	die "ERROR (XGrid.pm): Invalid row index. \n"; #debug
	}
else
	{
	$self->{'_table'}->idletasks;
	my $heading_width = $self->{'_row_headers'}[$row]->Width();
	$self->{'_row_headers'}[$row]->GeometryRequest($heading_width,$height);
	$self->{'_row_heights'}[$row] = $height;
	$self->{'_table'}->idletasks;
	}
}


=head3 delete_col

$object->delete_col($col);

Deletes the column from the table. All columns right of this column are moved one step to the left.

Parameters:

(+) $col: The column index

=cut

###################
# Delete a column #
###################
sub delete_col
{
my $self = shift;
my $col = shift;

#Test if the col exists
if ($col > $self->cols() or $col <1)
	{
	die "ERROR (XGrid.pm): Invalid column index. \n"; #debug
	}
else
	{
	#If a cell is currently activated, deactivate it
		if ( not $self->{'_active_cell_col'} == 0 or not $self->{'_active_cell_row'} == 0 )
			{
			$self->deactivate_cell($self->{'_active_cell_col'}, $self->{'_active_cell_row'});
			$self->{'_table'}->idletasks;
			}

	#If the col is not the last one, shift all cols right of it one step to the left
	if ($col < $self->cols() )
		{
		my $start_col = $col+1; 
		my $end_col = $self->cols();
		$self->_move_cols($start_col, $end_col, -1);
		}
	#Now delete the last column
	#Let's start with the header
	$self->{'_col_headers'}[$self->cols()]->gridForget();
	$self->{'_col_headers'}[$self->cols()]->destroy();
	$self->{'_col_widths'}[$self->cols()] = undef;
	#Iterate over all rows and remove the entries (Labels)
	for (my $y = 1; $y <= $self->rows(); $y++)
		{
		#Remove the Labels
		$self->{'_cell_entries'}[$self->cols()][$y]->gridForget();
		$self->{'_cell_entries'}[$self->cols()][$y]->destroy();
		#Remove all other stored info for the cell
		$self->{'_cell_contents'}[$self->cols()][$y] = undef;
		$self->{'_cell_styles'}[$self->cols()][$y] = undef;
		$self->{'_cell_contents'}[$self->cols()][$y] = undef;
		}
	#Decrease the number of columns
	$self->cols($self->cols()-1);
	#Redraw the whole thing
	$self->update();
	}
}

=head3 delete_row

$object->delete_row($row);

Deletes the row from the table. All rows below this row are moved one step upward.

Parameters:

(+) $row: The row index

=cut

################
# Delete a row #
################
sub delete_row
{
my $self = shift;
my $row = shift;

#Test if the row exists
if ($row > $self->rows() or $row <1)
	{
	die "ERROR (XGrid.pm): Invalid row index. \n"; #debug
	}
else
	{
	#If a cell is currently activated, deactivate it
		if ( not $self->{'_active_cell_col'} == 0 or not $self->{'_active_cell_row'} == 0 )
			{
			$self->deactivate_cell($self->{'_active_cell_col'}, $self->{'_active_cell_row'});
			$self->{'_table'}->idletasks;
			}

	#If the row is not the last one, shift all rows below it one step upwards
	if ($row < $self->rows() )
		{
		my $start_row = $row+1; 
		my $end_row = $self->rows();
		$self->_move_rows($start_row, $end_row, -1);
		}
	#Now delete the last column
	#Let's start with the header
	$self->{'_row_headers'}[$self->rows()]->gridForget();
	$self->{'_row_headers'}[$self->rows()]->destroy();
	$self->{'_row_heights'}[$self->rows()] = undef;
	#Iterate over all cols and remove the entries (Labels)
	for (my $x = 1; $x <= $self->cols(); $x++)
		{
		#Remove the Labels
		$self->{'_cell_entries'}[$x][$self->rows()]->gridForget();
		$self->{'_cell_entries'}[$x][$self->rows()]->destroy();
		#Remove all other stored info for the cell
		$self->{'_cell_contents'}[$x][$self->rows()] = undef;
		$self->{'_cell_styles'}[$x][$self->rows()] = undef;
		$self->{'_cell_contents'}[$x][$self->rows()] = undef;
		}
	#Decrease the number of row
	$self->rows($self->rows()-1);
	#Redraw the thing
	$self->update();
	}
}

=head3 insert_col

$object->insert_col($col_index);

Inserts a new column at the specified column index. The column that is there before
inserting and all columns below this row are moved one step to the right.

This method is slower than the method 'add_col' because of the managing
and moving of the existing columnss below the index.

Parameters:

(+) $col_index: The column index

=cut

###################
# Insert a column #
###################
sub insert_col
{
my $self = shift;
my $col = shift;

#Test if the row exists
if ($col > $self->cols() or $col < 1)
	{
	die "ERROR (XGrid.pm): Invalid col index in method 'insert_col'.\n"; #debug
	}
else
	{
	#If a cell is currently activated, deactivate it
		if ( not $self->{'_active_cell_col'} == 0 or not $self->{'_active_cell_row'} == 0 )
			{
			$self->deactivate_cell($self->{'_active_cell_col'}, $self->{'_active_cell_row'});
			$self->{'_table'}->idletasks;
			}

	#Now insert a new and clear column at the very right of the table
	$self->add_col($default_cell_width);

	#Move all columns one step downwards
	my $start_col = $col; 
	my $end_col = $self->cols()-1;
	$self->_move_cols($start_col, $end_col, +1);

	#Empty content and style of the inserted column (There is still the stuff from the old column at this index there)
	$self->clear_col($col, 1);

	#Size the newly inserted row
	$self->set_col_width($col, $default_cell_width);

	#Redraw the thing
	$self->update();
	}
}


=head3 insert_row

$object->insert_row($row_index);

Inserts a new row at the specified row index. The row that is there before
inserting and all rows below this row are moved one step downwards.

This method is slower than the method 'add_row' because of the managing
and moving of the existing rows below the index.

Parameters:

(+) $row_index: The row index

=cut

################
# Insert a row #
################
sub insert_row
{
my $self = shift;
my $row = shift;

#Test if the row exists
if ($row > $self->rows() or $row < 1)
	{
	die "ERROR (XGrid.pm): Invalid row index in method 'insert_row'.\n"; #debug
	}
else
	{
	#If a cell is currently activated, deactivate it
		if ( not $self->{'_active_cell_col'} == 0 or not $self->{'_active_cell_row'} == 0 )
			{
			$self->deactivate_cell($self->{'_active_cell_col'}, $self->{'_active_cell_row'});
			$self->{'_table'}->idletasks;
			}

	#Now insert a new and clear row at the bottom of the table
	$self->add_row($default_cell_height);

	#Move all rows one step downwards
	my $start_row = $row; 
	my $end_row = $self->rows()-1;
	$self->_move_rows($start_row, $end_row, +1);

	#Empty content and style of the inserted row (There is still the stuff from the old row at this index there)
	$self->clear_row($row, 1);


	#Size the newly inserted row
	$self->set_row_height($row, $default_cell_height);

	#Redraw the thing
	$self->update();
	}
}

=head3 clear_col

$object->clear_col($col, $do_clear_styles)

This method clears the content of the column. If $do_clear_styles is set to 1,
the stales are cleared too.

Parameters:

(+) $col: The index of the column (starting with 1, column 0 is reserved for the headers)

(+) $do_clear_styles: If 1, styles are cleared too

=cut

####################################
# Clearing the content of a column #
####################################
sub clear_col
{
my $self = shift;
my $col = shift;
my $do_clear_styles = shift;

$do_clear_styles = 0 if not defined $do_clear_styles;

if ($col > $self->cols() or $col < 1)
	{
	die "ERROR (XGrid.pm): Invalid column index. in method 'clear_col'.\n"; #debug
	}
else
	{
	$self->deactivate_cell(); 
	for (my $y = 1; $y <= $self->rows(); $y++)
		{
		#Remove the Labels
		$self->{'_cell_contents'}[$col][$y] = '';
		if ($do_clear_styles == 1)
			{
			$self->set_cell_style($col, $y, 'BASE');
			}
		}
	}
}

=head3 clear_row

$object->clear_row($row, $do_clear_styles)

This method clears the content of the row. If $do_clear_styles is set to 1,
the stales are cleared too.

Parameters:

(+) $row: The index of the row (starting with 1, row 0 is reserved for the headers)

(+) $do_clear_styles: If 1, styles are cleared too

=cut

#################################
# Clearing the content of a row #
#################################
sub clear_row
{
my $self = shift;
my $row = shift;
my $do_clear_styles = shift;

$do_clear_styles = 0 if not defined $do_clear_styles;

if ($row > $self->rows() or $row < 1)
	{
	die "ERROR (XGrid.pm): Invalid row index. in method 'clear_row'.\n"; #debug
	}
else
	{
	$self->deactivate_cell(); 
	for (my $x = 1; $x <= $self->cols(); $x++)
		{
		#Remove the Labels
		$self->{'_cell_contents'}[$x][$row] = '';
		if ($do_clear_styles == 1)
			{
			$self->set_cell_style($x, $row, 'BASE');
			}
		}
	}
}

=head3 set_col_style

$object->set_col_style($col, $style_name)

This method sets a style in a whole column.

Parameters:

(+) $col: The index of the column (starting with 1, column 0 is reserved for the headers)

(+) $style_name: The name of the style to be set

=cut

#################################
# Set a style in a whole column #
#################################
sub set_col_style
{
my $self = shift;
my $col = shift;
my $style_name = shift;

if ($col > $self->cols() or $col < 1)
	{
	die "ERROR (XGrid.pm): Invalid column index. in method 'set_col_style'.\n"; #debug
	}
else
	{
	$self->deactivate_cell(); 
	for (my $y = 1; $y <= $self->rows(); $y++)
		{
		$self->set_cell_style($col, $y, $style_name);
		}
	#Update the whole thing
	$self->{'_table'}->update();
	}
}

=head3 set_row_style

$object->set_row_style($row, $style_name)

This method sets a style in a whole row.

Parameters:

(+) $row: The index of the row (starting with 1, row 0 is reserved for the headers)

(+) $style_name: The name of the style to be set

=cut

##############################
# Set a style in a whole row #
##############################
sub set_row_style
{
my $self = shift;
my $row = shift;
my $style_name = shift;

if ($row > $self->rows() or $row < 1)
	{
	die "ERROR (XGrid.pm): Invalid row index. in method 'set_row_style'.\n"; #debug
	}
else
	{
	$self->deactivate_cell(); 
	for (my $x = 1; $x <= $self->cols(); $x++)
		{
		$self->set_cell_style($x, $row, $style_name);
		}
	#Update the whole thing
	$self->{'_table'}->update();
	}
}

=head3 set_table_style

$object->set_table_style($style_name)

This method sets a style in the whole table.

Parameters:

(+) $style_name: The name of the style to be set

=cut

##################################
# Set a style in the whole table #
##################################
sub set_table_style
{
my $self = shift;
my $style_name = shift;

$self->deactivate_cell(); 
for (my $y = 1; $y <= $self->rows(); $y++)
	{
	for (my $x = 1; $x <= $self->cols(); $x++)
		{
		$self->set_cell_style($x, $y, $style_name);
		}
	#Update the whole thing
	$self->{'_table'}->update();
	}
}


=head2 Private methods:

DO NOT USE THESE METHODS !

This methods are described here for the use of widget developer, who plan to extend
this class or want to derive from it.

=head3 _create_cell

$ref_widget = $object->_create_cell($col, $row [,$col_width, $row_height, $do_grid] ); 

This function creates a cell with an embeded Text widget in the table

The function returns a reference to that widget.

Parameters:

(+) $col: The index of the col (starting with 1, column 0 is reserved for the headers)

(+) $row: The index of the row (starting with 1, row 0 is reserved for the headers)

(+) $col_width: The width of the column (OPTIONAL, if undef, the width will be taken from the width of the column header)

(+) $row_height: The height of the row (OPTIONAL, if undef, the height will be taken from the width of the row header)

(+) $do_grid: 1 if the cell should be automatically gridded, 0 or undef if this should not happen automatically

=cut

###############################################
# Adding a new row at the bottom of the table #
###############################################
sub _create_cell
{
my $self = shift;
my $col = shift;
my $row = shift;
my $col_width = shift;
my $row_height = shift;
my $do_grid = shift;

#Cretae the entry
my $cell_height;
my $cell_width;

if (not defined $col_width)
	{
	$cell_width = $self->{'_col_widths'}[$col];
	}
else
	{
	$cell_width = $col_width;
	}

if (not defined $row_height)
	{
	$cell_height = $self->{'_row_heights'}[$row];
	}
else
	{
	$cell_height = $row_height;
	}

my $entry = $self->{'_cell_entries'}[$col][$row] = $self->{'_table'}->Label( -width=>1, -height=>1, -textvariable=>\$self->{'_cell_contents'}[$col][$row] );
$self->{'_cell_styles'}[$col][$row] = 'BASE';
$self->set_cell_style($col, $row, 'BASE');
$entry->bind('<Button-3>' => [\&_cell_popup, Ev('X'), Ev('Y'), $self, $col, $row]);

#Key-bindings for the Entry
$entry->bind('<Control-Key-Down>', [\&_nav_key_released, $self, 'down', $row, $col]);
$entry->bind('<Control-Key-Up>', [\&_nav_key_released, $self, 'up', $row, $col]);
$entry->bind('<Control-Key-Right>', [\&_nav_key_released, $self, 'right', $row, $col]);
$entry->bind('<Control-Key-Left>', [\&_nav_key_released, $self, 'left', $row, $col]);
$entry->bind('<Control-Key-Home>', sub{$self->{'_table'}->see($self->{'_table_button'})});
$entry->bind('<Control-Key-Home>', sub{$self->_show_left_upper_cell() } );
$entry->bind('<Control-Key-End>', sub{$self->_show_right_bottom_cell() } );
$entry->bind('<Key>', [\&_activate_cell, $self, $col, $row]);
$entry->bind('<Button>', [\&_activate_cell, $self, $col, $row]);

if (defined $do_grid and $do_grid == 1)
	{
	$entry->grid(-row=>$row, -column=>$col,  -sticky=>'nsew');
	}

return $entry;
}

=head3 _nav_key_released

$object->_nav_key_released(); 

This is the callback for cursor keys, 'tab', 'return'/'enter', 'home' and 'end' keys.

=cut

###########################
# Navigation keys relesed #
###########################
sub _nav_key_released
{
my $sender = shift;
my $self = shift;
my $key = shift; 

if (not $self->{'_active_cell_col'} == 0 and not $self->{'_active_cell_row'} == 0)
	{
	my $do_jump = 0; #flag
	my $x = $self->{'_active_cell_col'};
	my $y = $self->{'_active_cell_row'};
	if ($key eq 'down')
		{
		$y++;
		$y = 1 if $y > $self->rows();
		$do_jump = 1; #setting the flag that a jump should be performed
		}
	elsif ($key eq 'up')
		{
		$y--;
		$y = $self->rows() if $y == 0;
		$do_jump = 1; #setting the flag that a jump should be performed
		}
	elsif ($key eq 'right') 
		{
		$x++;
		$x = 1 if $x > $self->cols();
		$do_jump = 1; #setting the flag that a jump should be performed
		}
	elsif ($key eq 'left')
		{
		$x--;
		$x = $self->cols() if $x == 0;
		$do_jump = 1; #setting the flag that a jump should be performed
		}
	
	if ($do_jump == 1)
		{ 
		_activate_cell(undef, $self, $x, $y);
		$self->{'_table'}->idletasks;
		if ($x==1 and $y == 1 )
			{
			$self->{'_table'}->see($self->{'_table_button'});
		   }
		elsif ($y == 1 )
			{
			$self->{'_table'}->see($self->{'_col_headers'}[$x]);
		   }
		elsif ($x == 1)
		 	{
		 	$self->{'_table'}->see($self->{'_row_headers'}[$y]);
			}
		else
			{
			$self->{'_table'}->see($self->{'_text_entry'});
			}
		}
	}
}

=head3 _col_heading_mouse_button1_pressed

$object->_col_heading_mouse_button1_pressed(); 

This is the callback for pressing the left mousebutton on a column-heading

=cut

#######################################
# MousButton pressed on column header #
#######################################
sub _col_heading_mouse_button1_pressed
{
my $sender = shift;
my $self = shift;
my $X = shift;
my $Y = shift;
my $col = shift;

$self->{'_col_drag'} = 1;
$self->{'_col_drag_X'} = $X;
$sender->configure(-cursor=>'sb_h_double_arrow');
}

=head3 _col_heading_mouse_button1_released

$object->_col_heading_mouse_button1_released(); 

This is the callback for releasing the left mousebutton on a column-heading

=cut

########################################
# MousButton released on column header #
########################################
sub _col_heading_mouse_button1_released
{
my $sender = shift;
my $self = shift;
my $X = shift;
my $Y = shift;
my $col = shift;

if (exists $self->{'_col_drag'} and defined $self->{'_col_drag'} and $self->{'_col_drag'}==1)
	{
	$self->{'_col_drag'} = 0;
	$self->{'_table'}->idletasks;
	my $incr_x = $X - $self->{'_col_drag_X'};
	my $col_width = $self->{'_col_widths'}[$col];
	my $heading_height = $self->{'_col_headers'}[$col]->Height();
	$col_width += $incr_x;
	$self->{'_col_headers'}[$col]->GeometryRequest($col_width,$heading_height);
	$self->{'_col_widths'}[$col] = $col_width;
	$sender->configure(-cursor=>'hand2');
	}
$self->{'_table'}->idletasks;
}

=head3 _col_heading_popup

This is the callback for presing the right mouse button on a column heading.

This displays the column-popup menu.

=cut

################
# Column-Popup #
################
sub _col_heading_popup
{
my $sender = shift;
my $X = shift;
my $Y = shift;
my $self = shift;
my $col = shift;

#POPUP-Menu creation
my $menu= $self->Menu(-tearoff=>0);
$menu->add ('command', -label=>'Delete this column', -command => sub{$self->delete_col($col)} );
$menu->add ('command', -label=>'Insert a new column', -command => sub{$self->insert_col($col)} );
$menu->add ('command', -label=>'Empty this column (Content)', -command => sub{$self->clear_col($col, 0)} );
$menu->add ('command', -label=>'Empty this column (Content & styles)', -command => sub{$self->clear_col($col, 1)} );

#Create the style menu
my $style_menu = $menu->menu->Menu(-tearoff=>0);
foreach my $style ($self->style_list())
	{
	$style_menu->add ('command', -label=>$style, -command=>sub{ $self->set_col_style($col, $style) });
	}
$menu->cascade(-label=>'Styles');
$menu->entryconfigure('Styles', -menu=>$style_menu);


#Showing it
$menu->Popup(-popover=>'cursor', -popanchor=>'se');
}

=head3 _row_heading_mouse_button1_pressed

$object->_row_heading_mouse_button1_pressed(); 

This is the callback for pressing the left mousebutton on a row-heading

=cut

####################################
# MousButton pressed on row header #
####################################
sub _row_heading_mouse_button1_pressed
{
my $sender = shift;
my $self = shift;
my $X = shift;
my $Y = shift;
my $row = shift;

$self->{'_row_drag'} = 1;
$self->{'_row_drag_Y'} = $Y;
$sender->configure(-cursor=>'sb_v_double_arrow');
}

=head3 _row_heading_mouse_button1_released

$object->_row_heading_mouse_button1_released(); 

This is the callback for releasing the left mousebutton on a row-heading

=cut

#####################################
# MousButton released on row header #
#####################################
sub _row_heading_mouse_button1_released
{
my $sender = shift;
my $self = shift;
my $X = shift;
my $Y = shift;
my $row = shift;

if (exists $self->{'_row_drag'} and defined $self->{'_row_drag'} and $self->{'_row_drag'}==1)
	{
	$self->{'_row_drag'} = 0;
	$self->{'_table'}->idletasks;
	my $incr_y = $Y - $self->{'_row_drag_Y'};
	my $row_height = $self->{'_row_heights'}[$row];
	my $heading_width = $self->{'_row_headers'}[$row]->Width();
	$row_height += $incr_y;
	$self->{'_row_headers'}[$row]->GeometryRequest($heading_width, $row_height);
	$self->{'_row_heights'}[$row] = $row_height;
	$self->{'_table'}->idletasks;
	$sender->configure(-cursor=>'hand2');
	}
}

=head3 _row_heading_popup

This is the callback for presing the right mouse button on a row heading.

This displays the row-popup menu.

=cut

################
# Column-Popup #
################
sub _row_heading_popup
{
my $sender = shift;
my $X = shift;
my $Y = shift;
my $self = shift;
my $row = shift;


#POPUP-Menu creation
my $menu= $self->Menu(-tearoff=>0);
$menu->add ('command', -label=>'Delete this row', -command => sub{$self->delete_row($row)} );
$menu->add ('command', -label=>'Insert a new row', -command => sub{$self->insert_row($row)} );
$menu->add ('command', -label=>'Empty this row (Content)', -command => sub{$self->clear_row($row, 0)} );
$menu->add ('command', -label=>'Empty this row (Content & styles)', -command => sub{$self->clear_row($row, 1)} );

#Create the style menu
my $style_menu = $menu->menu->Menu(-tearoff=>0);
foreach my $style ($self->style_list())
	{
	$style_menu->add ('command', -label=>$style, -command=>sub{ $self->set_row_style($row, $style) });
	}
$menu->cascade(-label=>'Styles');
$menu->entryconfigure('Styles', -menu=>$style_menu);

#Showing it
$menu->Popup(-popover=>'cursor', -popanchor=>'se');

}

=head3 _cell_popup

This is the callback for presing the right mouse button on a cell

This displays the cell-popup menu.

=cut

##############
# Cell-Popup #
##############
sub _cell_popup
{
my $sender = shift;
my $X = shift;
my $Y = shift;
my $self = shift;
my $col = shift;
my $row = shift;

#POPUP-Menu creation
my $menu = $self->Menu(-tearoff=>0);

#Create style menu
my $style_menu = $menu->menu->Menu(-tearoff=>0);
foreach my $style ($self->style_list())
	{
	$style_menu->add ('command', -label=>$style, -command=>sub{ $self->set_cell_style($col, $row, $style) });
	}
$menu->cascade(-label=>'Styles');
$menu->entryconfigure('Styles', -menu=>$style_menu);

#Showing it
$menu->Popup(-popover=>'cursor', -popanchor=>'se');
}

=head3 _table_popup

This is the callback for presing the right mouse button on the 'table'-button (the one on the upper left).

This displays the table-popup menu.

=cut

###############
# Table-Popup #
###############
sub _table_popup
{
my $sender = shift;
my $X = shift;
my $Y = shift;
my $self = shift;

#POPUP-Menu creation
my $menu = $self->Menu(-tearoff=>0);
$menu->add ('command', -label=>'Add one row', -command=>sub{ $self->add_row($default_cell_height) });
$menu->add ('command', -label=>'Add one column', -command=>sub{ $self->add_col($default_cell_width) });
$menu->add ('command', -label=>'Empty table (Content)', -command => sub{
																								for (my $col = 1; $col <= $self->cols(); $col++)
																									{
																									$self->clear_col($col, 0);
																									}
																								} );
$menu->add ('command', -label=>'Empty table (Content & styles)', -command => sub{
																								for (my $col = 1; $col <= $self->cols(); $col++)
																									{
																									$self->clear_col($col, 1);
																									}
																								} );

#Create the style menu
my $style_menu = $menu->menu->Menu(-tearoff=>0);
foreach my $style ($self->style_list())
	{
	$style_menu->add ('command', -label=>$style, -command=>sub{ $self->set_table_style($style) });
	}
$menu->cascade(-label=>'Styles');
$menu->entryconfigure('Styles', -menu=>$style_menu);

#Showing it
$menu->Popup(-popover=>'cursor', -popanchor=>'se');
}

=head3 _create_balloonhelp

$ref_baloon = _create_balloonhelp($ref_widget, $text, $status, $ref_statusbar);

Creates a balloonhelp and associates it with the widget

The functions returns a reference to the created balloon widget.

Parameters: 

(+) $ref_widget: A reference to the widget

(+) $text: The text do be displayed in the balloon

(+) $state: The state of the balloon ('balloon', 'status', 'both', 'none'). Default is 'balloon'

(+) $ref_statusbar: A reference for the statusbar(Label) that should display the balloonhelp's text, if state is 'status' or 'both', else set it to undef

=cut

######################################
# Creates a balloonhelp for a widget #
######################################
sub _create_balloonhelp
{
my $self = shift;
my $widget = shift;
my $text = shift;
my $state = shift;
my $ref_statusbar = shift;
#Check if all parameters are right
if (defined $widget)
	{
	$state = 'balloon' if not defined $state;
	if (not defined $ref_statusbar)
		{
		$state = 'balloon' if ($state eq 'status' or $state eq 'both');
		}
	#Create ballon
	my $balloon;
	if ($state eq 'balloon' or $state eq 'none')
		{
		$balloon = $widget->Balloon(-state=>$state);
		}
	else
		{
		$balloon = $widget->Balloon(-state=>$state, -statusbar=>$ref_statusbar);
		}
	$balloon->attach($widget, -msg=>$text);
	return $balloon;
	}
else
	{
	return undef;
	}
}

=head3 _activate_cell

$object->_activate_cell($col, $row);

Is called when a cell gets activated by user entry (Key or Mouse)

Parameters: 

(+) $col: The column that contains the cell

(+) $row: The row that contains the cell

=cut

###################
# Activate a cell #
###################
sub _activate_cell
{
my $sender = shift;
my $self = shift;
my $col = shift;
my $row = shift;
if (not $col == 0 or not $row == 0)
	{
	if (not $self->{'_active_cell_col'} == 0 and not $self->{'_active_cell_row'} == 0)
		{
		_deactivate_cell(undef, $self, $self->{'_active_cell_col'}, $self->{'_active_cell_row'});
		}
	$self->{'_text_entry'}->insert('1.0', $self->{'_cell_contents'}[$col][$row] );
	$self->{'_cell_entries'}[$col][$row]->gridForget();
	$self->{'_text_entry'}->GeometryRequest(4,4);
	$self->_configure_text_entry($self->{'_cell_styles'}[$col][$row]);
	$self->{'_text_entry'}->grid(-column=>$col, -row=>$row, -sticky=>'nsew');
	$self->{'_text_entry'}->focus();
	$self->{'_text_entry'}->bind('<FocusOut>',[\&_deactivate_cell, $self, $col, $row]);
	$self->{'_active_cell_col'} = $col;
	$self->{'_active_cell_row'} = $row;
	}
}

=head3 _deactivate_cell

$object->_deactivate_cell($col, $row);

Is called when a cell gets deactivated by user entry (Key or Mouse)

Parameters: 

(+) $col: The column that contains the cell

(+) $row: The row that contains the cell

=cut

#####################
# Deactivate a cell #
#####################
sub _deactivate_cell
{
my $sender = shift;
my $self = shift;
my $col = shift;
my $row = shift;

if (not $col == 0 or not $row == 0)
	{
	my $content = $self->{'_text_entry'}->get('1.0', 'end');
	chomp($content) ;
	$self->{'_cell_contents'}[$col][$row] =  $content;
	$self->{'_text_entry'}->delete('1.0', 'end');
	$self->{'_text_entry'}->gridForget();
	$self->_configure_label_entry($col, $row, $self->{'_cell_styles'}[$col][$row]);
	$self->{'_cell_entries'}[$col][$row]->grid(-column=>$col, -row=>$row, -sticky=>'nsew');
	$self->{'_active_cell_col'} = 0;
	$self->{'_active_cell_row'} = 0;
	$self->{'_table'}->idletasks;
	}
}

=head3 _test_cell

$object->_test_cell($col, $row, $calling_sub);

Test if the cell exists and print an error message if not.

The function returns 1 if the cell exists or undef if not.

Parameters: 

(+) $col: The column that contains the cell

(+) $row: The row that contains the cell

(+) $calling_sub: The name of the function that called (For the error message)

=cut

############################
# Test a cell if it exists #
############################
sub _test_cell
{
my $self = shift;
my $col = shift;
my $row = shift;
my $calling_sub = shift;

my $result = undef ;

if ( $col >= 1 and $col <= $self->cols() and $row >= 1 and $row <= $self->rows() )
	{
	$result = 1;
	}
else
	{
	die ("ERROR in XGrid.pm: Wrong column or row index in method '$calling_sub'. \n"); #debug
	}

return $result;
}

=head3 _configure_text_entry

$object->_configure_text_entry($style_name);

Change the display style of the Text widget


Parameters: 

(+) $style_name: The name of the style to be set

=cut

#############################
# Configure the Text widget #
#############################
sub _configure_text_entry
{
my $self = shift;
my $style_name = shift;

my %text_widget_options = %{$self->{'_style_defs'}->{$style_name}->{'text_widget'}} ;
$self->{'_text_entry'}->configure(%text_widget_options);

foreach my $tag (keys %{$self->{'_style_defs'}})
	{
	$self->{'_text_entry'}->tagRemove($tag, '1.0', 'end');
	}

$self->{'_text_entry'}->tagAdd($style_name, '1.0', 'end');
}

=head3 _configure_label_entry

$object->_configure_label_entry($col, $row, $style_name);

Change the display style of the cell. (The label there).


Parameters: 

(+) $col: The column that contains the cell

(+) $row: The row that contains the cell

(+) $style_name: The name of the style to be set

=cut

############################
# Configure a Label widget #
############################
sub _configure_label_entry
{
my $self = shift;
my $col = shift;
my $row = shift;
my $style_name = shift;

if (defined $self->_test_cell($col, $row, '_configure_label_entry') )
	{
	my %label_widget_options = %{$self->{'_style_defs'}->{$style_name}->{'label_widget'}} ;
	$self->{'_cell_entries'}[$col][$row]->configure(%label_widget_options);
	}

}

=head3 _show_left_upper_cell

$object->_show_left_upper_cell();

Changes the display so that the cell in the very left column and in the first row is displayed.

=cut

###############################
# Jump to the left-upper cell #
###############################
sub _show_left_upper_cell
{
my $self = shift;

if ($self->cols() > 0 and $self->rows() > 0)
	{
	$self->{'_table'}->see($self->{'_table_button'});
	$self->activate_cell(1,1);
	}
}

=head3 _show_right_bottom_cell

$object->_show_right_bottom_cell();

Changes the display so that the cell in the very right column and in the last row is displayed.

=cut

#################################
# Jump to the right-bottom cell #
#################################
sub _show_right_bottom_cell
{
my $self = shift;

if ($self->cols() > 0 and $self->rows() > 0)
	{
	$self->{'_table'}->see($self->{'_cell_entries'}[$self->cols()][$self->rows()]);
	$self->activate_cell($self->cols(), $self->rows());
	}
}


=head3 _move_cols

$object->_move_cols($start_col, $end_col, $offset);

Moves the ranges specified by $start_col and $end_col.


Parameters: 

(+) $start_col: The left column in the range to be moved

(+) $end_col: The right column in the range to be moved

(+) $offset: How many cols the range should be moved. Positive values
shift to the right, negative values shift to the left. The range is not allowed
to exceed the borders of the table, otherwise it gives an ERROR.

This means that $offset has to be more than ($start_col * -1) and
less than ($self->cols() - $end_col)

=cut

##################
# Moving columns #
##################
sub _move_cols
{
my $self = shift;
my $start_col = shift;
my $end_col = shift;
my $offset = shift;

goto EXIT if $offset == 0;

#Test ranges
if ($start_col < 1 or $start_col > $self->cols() or $end_col < 1 or $end_col > $self->cols() )
	{
	die "ERROR (XGrid.pm): Invalid column index in method '_move_cols'.\n";
	goto EXIT;
	}
if ($offset < ($start_col * -1) or $offset > ($self->cols() - $end_col) )
	{
	die "ERROR (XGrid.pm): Invalid offset in method '_move_cols'.\n";
	goto EXIT;
	}

# Flip them if they are in the wrong order
if ($start_col > $end_col)
	{
	my $buffer = $start_col;
	$start_col = $end_col;
	$end_col = $buffer;
	}

my $iterator;
my $start_value;
my $end_value;
if ($offset > 0) 
	{
	$iterator = -1;
	$start_value = $end_col;
	$end_value = $start_col;
	} 
else 
	{
	$iterator = +1;
	$start_value = $start_col;
	$end_value = $end_col;
	}

my $col = $start_value;
#Copy the vlaues, styles and columnwidths
while ($col >= $start_col and $col <= $end_col)
	{
	#Copy column width
	$self->set_col_width($col+$offset, $self->get_col_width($col) );

	#Copy contents and styles
	for (my $row = 1; $row <= $self->rows(); $row++)
		{
		$self->{'_cell_contents'}[$col+$offset][$row] = $self->{'_cell_contents'}[$col][$row] ;
		$self->set_cell_style($col+$offset, $row, $self->{'_cell_styles'}[$col][$row] );
		}
	$col = $col + $iterator;
	}

EXIT:
}

=head3 _move_rows

$object->_move_rows($start_row, $end_row, $offset);

Moves the ranges specified by $start_row and $end_row.


Parameters: 

(+) $start_row: The upper row in the range to be moved

(+) $end_row: The lower row in the range to be moved

(+) $offset: How many rows the range should be moved. Positive values
shift downwards, negative values shift upwards. The range is not allowed
to exceed the borders of the table, otherwise it gives an ERROR.

This means that $offset has to be more than ($start_row * -1) and
less than ($self->rows() - $end_row)

=cut

###############
# Moving rows #
###############
sub _move_rows
{
my $self = shift;
my $start_row = shift;
my $end_row = shift;
my $offset = shift;

goto EXIT if $offset == 0;

#Test ranges
if ($start_row < 1 or $start_row > $self->rows() or $end_row < 1 or $end_row > $self->rows() )
	{
	die "ERROR (XGrid.pm): Invalid row index in method '_move_rows'.\n";
	goto EXIT;
	}
if ($offset < ($start_row * -1) or $offset > ($self->rows() - $end_row) )
	{
	die "ERROR (XGrid.pm): Invalid offset in method '_move_rows'.\n";
	goto EXIT;
	}

# Flip them if they are in the wrong order
if ($start_row > $end_row)
	{
	my $buffer = $start_row;
	$start_row = $end_row;
	$end_row = $buffer;
	}

my $iterator;
my $start_value;
my $end_value;
if ($offset > 0) 
	{
	$iterator = -1;
	$start_value = $end_row;
	$end_value = $start_row;
	} 
else 
	{
	$iterator = +1;
	$start_value = $start_row;
	$end_value = $end_row;
	}

my $row = $start_value;
#Copy the vlaues, styles and rowheights
while ($row >= $start_row and $row <= $end_row)
	{
	#Copy row height
	$self->set_row_height($row+$offset, $self->get_row_height($row) );

	#Copy contents and styles
	for (my $col = 1; $col <= $self->cols(); $col++)
		{
		$self->{'_cell_contents'}[$col][$row+$offset] = $self->{'_cell_contents'}[$col][$row] ;
		$self->set_cell_style($col, $row+$offset, $self->{'_cell_styles'}[$col][$row] );
		}
	$row = $row + $iterator;
	}

EXIT:
}


1; #Thats the way a package should end.

=head1 BINDINGS:

Ctrl + Up = Move to the cell above the active one cell.

Ctrl + Down = Move to the cell under the atcive one  cell.

Ctrl + Left = Move to the cell left of the atcive one  cell.

Ctrl + Right = Move to the cell right of the atcive one  cell.

Ctrl + Home = Move to the upper left corner of the table

Ctrl + End = Move to the lower right corner of the table

=head1 KNOWN BUGS:

This widget is to slow in creating the cells !

Everything more than a table of 12x30 cells on a Celeron366 is beyond my patience.

=head1 REQUIREMENTS:

Nothing except Perl and perl/Tk (should work with ptk 800.016+).

=head1 STILL MISSING:

(+) Speed improvements (especially in building up the cols/rows).

(+) Possibility to select a range of cells.

(+) Do not 'hardcode' the default cell_width and cell_height that is used, define two new options for that.

(+) Clipboard functionality that copies the selected range (including cell-styles of course).

(+) Adding more options for Populate (Examples: Turning on/off of the headers [-showheaders=>0/1], 
the background- and foregroundcolor of the selected cell [-activecellbackground, -activecellforeground], 
allowing to specify other images for the resizing grips [-hgripimage, -vgripimage, -tableimage], ... ).

(+) Doing all of this 'ConfigSpecs', 'Adveritse', 'Delegate stuff'

(+) Allowing to add items+callbacks to the table-, column-, row- and cell-menus , to replace or to hide them totally.

(+) Loading from comma- , semicolon- and tab-separated lists and saving into these formats too.

(+) Column- and rowspan for cells

(+) Writing a better doc.

=head1 LICENSE:

This widget is licensed the same way as Perl itself.

=head1 AUTHOR:

Copyright by Thomas Schmickl, Graz, Austria, email: schmickl@nextra.at, faultier@nextra.at or edoc@nextra.at.

If these emailadresses are expired some day, try to contact me via: office@foto-schmickl.at and 
write "For Thomas Schmickl:" into the subject line.

=cut
